@@ -43,12 +43,11 @@
 
 #include <set>
 
+#include "lnav_util.hh"
 #include "line_buffer.hh"
 
 using namespace std;
 
-static const size_t DEFAULT_LINE_BUFFER_SIZE   = 256 * 1024;
-static const size_t MAX_LINE_BUFFER_SIZE       = 4 * DEFAULT_LINE_BUFFER_SIZE;
 static const size_t DEFAULT_INCREMENT          = 128 * 1024;
 static const size_t MAX_COMPRESSED_BUFFER_SIZE = 32 * 1024 * 1024;
 
@@ -103,18 +102,18 @@ private:
 
         snprintf(lockname, sizeof(lockname), "/tmp/lnav.%d.lck", getpid());
         this->lh_fd = open(lockname, O_CREAT | O_RDWR, 0600);
-        fcntl(this->lh_fd, F_SETFD, FD_CLOEXEC);
+        log_perror(fcntl(this->lh_fd, F_SETFD, FD_CLOEXEC));
         unlink(lockname);
     };
 
-    int lh_fd;
+    auto_fd lh_fd;
 };
 /* XXX END */
 
 line_buffer::line_buffer()
     : lb_gz_file(NULL),
       lb_bz_file(false),
-      lb_gz_offset(0),
+      lb_compressed_offset(0),
       lb_file_size((size_t)-1),
       lb_file_offset(0),
       lb_file_time(0),
@@ -172,9 +171,13 @@ throw (error)
                 if (gz_id[0] == '\037' && gz_id[1] == '\213') {
                     int gzfd = dup(fd);
 
-                    fcntl(gzfd, F_SETFD, FD_CLOEXEC);
-                    lseek(fd, 0, SEEK_SET);
+                    log_perror(fcntl(gzfd, F_SETFD, FD_CLOEXEC));
+                    if (lseek(fd, 0, SEEK_SET) < 0) {
+                        close(gzfd);
+                        throw error(errno);
+                    }
                     if ((this->lb_gz_file = gzdopen(gzfd, "r")) == NULL) {
+                        close(gzfd);
                         if (errno == 0) {
                             throw bad_alloc();
                         }
@@ -182,14 +185,18 @@ throw (error)
                             throw error(errno);
                         }
                     }
-                    this->lb_file_time = *((int32_t *)&gz_id[4]);
-                    if (this->lb_file_time < 0)
+                    this->lb_file_time = read_le32(
+                        (const unsigned char *)&gz_id[4]);
+                    if (this->lb_file_time < 0) {
                         this->lb_file_time = 0;
-                    this->lb_gz_offset = lseek(this->lb_fd, 0, SEEK_CUR);
+                    }
+                    this->lb_compressed_offset = lseek(this->lb_fd, 0, SEEK_CUR);
                 }
 #ifdef HAVE_BZLIB_H
                 else if (gz_id[0] == 'B' && gz_id[1] == 'Z') {
-                    lseek(fd, 0, SEEK_SET);
+                    if (lseek(fd, 0, SEEK_SET) < 0) {
+                        throw error(errno);
+                    }
                     this->lb_bz_file = true;
 
                     /*
@@ -197,6 +204,8 @@ throw (error)
                      * to keep as much in memory as possible.
                      */
                     this->resize_buffer(MAX_COMPRESSED_BUFFER_SIZE);
+
+                    this->lb_compressed_offset = 0;
                 }
 #endif
             }
@@ -220,6 +229,7 @@ throw (error)
 
     /* Still need more space, try a realloc. */
     old = this->lb_buffer.release();
+    this->lb_share_manager.invalidate_refs();
     tmp = (char *)realloc(old, new_max);
     if (tmp != NULL) {
         this->lb_buffer     = tmp;
@@ -249,9 +259,10 @@ throw (error)
          * The request is outside the cached range, need to reload the
          * whole thing.
          */
+        this->lb_share_manager.invalidate_refs();
         prefill = 0;
         this->lb_buffer_size = 0;
-        if ((this->lb_file_size != (size_t)-1) &&
+        if ((this->lb_file_size != (ssize_t)-1) &&
             (start + this->lb_buffer_max > this->lb_file_size)) {
             /*
              * If the start is near the end of the file, move the offset back a
@@ -272,16 +283,18 @@ throw (error)
         prefill = start - this->lb_file_offset;
     }
     require(this->lb_file_offset <= start);
-    require(prefill <= this->lb_buffer_size);
+    require(prefill <= (size_t)this->lb_buffer_size);
 
     available = this->lb_buffer_max - (start - this->lb_file_offset);
-    require(available <= this->lb_buffer_max);
+    require(available <= (size_t)this->lb_buffer_max);
 
     if (max_length > available) {
         /*
          * Need more space, move any existing data to the front of the
          * buffer.
          */
+        this->lb_share_manager.invalidate_refs();
+
         this->lb_buffer_size -= prefill;
         this->lb_file_offset += prefill;
         memmove(&this->lb_buffer[0],
@@ -293,8 +306,6 @@ throw (error)
             this->resize_buffer(this->lb_buffer_max +
                                 DEFAULT_LINE_BUFFER_SIZE);
         }
-
-        this->lb_share_manager.invalidate_refs();
     }
 }
 
@@ -317,7 +328,7 @@ throw (error)
 
         /* ... read in the new data. */
         if (this->lb_gz_file) {
-            if (this->lb_file_size != (size_t)-1 &&
+            if (this->lb_file_size != (ssize_t)-1 &&
                 this->in_range(start) &&
                 this->in_range(this->lb_file_size - 1)) {
                 rc = 0;
@@ -325,20 +336,25 @@ throw (error)
             else {
                 lock_hack::guard guard;
 
-                lseek(this->lb_fd, this->lb_gz_offset, SEEK_SET);
+                lseek(this->lb_fd, this->lb_compressed_offset, SEEK_SET);
                 gzseek(this->lb_gz_file,
                        this->lb_file_offset + this->lb_buffer_size,
                        SEEK_SET);
                 rc = gzread(this->lb_gz_file,
                             &this->lb_buffer[this->lb_buffer_size],
                             this->lb_buffer_max - this->lb_buffer_size);
-                this->lb_gz_offset = lseek(this->lb_fd, 0, SEEK_CUR);
+                this->lb_compressed_offset = lseek(this->lb_fd, 0, SEEK_CUR);
+                if (rc != -1 && (
+                        rc < (this->lb_buffer_max - this->lb_buffer_size))) {
+                    this->lb_file_size = (
+                            this->lb_file_offset + this->lb_buffer_size + rc);
+                }
             }
         }
 #ifdef HAVE_BZLIB_H
         else if (this->lb_bz_file) {
-            if (this->lb_file_size != (size_t)-1 &&
-                (((size_t)start >= this->lb_file_size) ||
+            if (this->lb_file_size != (ssize_t)-1 &&
+                (((ssize_t)start >= this->lb_file_size) ||
                  (this->in_range(start) &&
                   this->in_range(this->lb_file_size - 1)))) {
                 rc = 0;
@@ -348,13 +364,19 @@ throw (error)
                 char             scratch[32 * 1024];
                 BZFILE *         bz_file;
                 off_t            seek_to;
+                int              bzfd;
 
                 /*
                  * Unfortunately, there is no bzseek, so we need to reopen the
                  * file every time we want to do a read.
                  */
-                lseek(this->lb_fd, 0, SEEK_SET);
-                if ((bz_file = BZ2_bzdopen(dup(this->lb_fd), "r")) == NULL) {
+                bzfd = dup(this->lb_fd);
+                if (lseek(this->lb_fd, 0, SEEK_SET) < 0) {
+                    close(bzfd);
+                    throw error(errno);
+                }
+                if ((bz_file = BZ2_bzdopen(bzfd, "r")) == NULL) {
+                    close(bzfd);
                     if (errno == 0) {
                         throw bad_alloc();
                     }
@@ -376,6 +398,7 @@ throw (error)
                 rc = BZ2_bzread(bz_file,
                                 &this->lb_buffer[this->lb_buffer_size],
                                 this->lb_buffer_max - this->lb_buffer_size);
+                this->lb_compressed_offset = lseek(bzfd, 0, SEEK_SET);
                 BZ2_bzclose(bz_file);
 
                 if (rc != -1 && (
@@ -485,7 +508,7 @@ throw (error)
             (request_size == MAX_LINE_BUFFER_SIZE) ||
             ((request_size > lv.lv_len) && lv.lv_len > 0)) {
             if ((lf != NULL) &&
-                ((lf - line_start) >= MAX_LINE_BUFFER_SIZE - 1)) {
+                ((size_t) (lf - line_start) >= MAX_LINE_BUFFER_SIZE - 1)) {
                 lf = NULL;
             }
             if (lf != NULL) {
@@ -503,6 +526,8 @@ throw (error)
             }
             else {
                 if (lv.lv_len >= MAX_LINE_BUFFER_SIZE) {
+                    log_warning("Line exceeded max size: offset=%d",
+                                offset);
                     lv.lv_len = MAX_LINE_BUFFER_SIZE - 1;
                     lv.lv_partial = false;
                 }
@@ -550,7 +575,7 @@ throw (error)
         }
     }
 
-    ensure(lv.lv_len <= this->lb_buffer_size);
+    ensure(lv.lv_len <= (size_t)this->lb_buffer_size);
     ensure(!retval ||
            (lv.lv_start >= this->lb_buffer &&
             (lv.lv_start + lv.lv_len) <= (this->lb_buffer + this->lb_buffer_size)));
@@ -559,17 +584,21 @@ throw (error)
     return retval;
 }
 
-bool line_buffer::read_line(off_t &offset_inout, shared_buffer_ref &sbr)
+bool line_buffer::read_line(off_t &offset_inout, shared_buffer_ref &sbr, line_value *lv)
     throw (error)
 {
-    line_value lv;
+    line_value lv_tmp;
     bool retval;
 
+    if (lv == NULL) {
+        lv = &lv_tmp;
+    }
+
     // Clear the incoming ref right away so that an invalidate
     // does not cause a wasted malloc/copy.
     sbr.disown();
-    if ((retval = this->read_line(offset_inout, lv))) {
-        sbr.share(this->lb_share_manager, lv.lv_start, lv.lv_len);
+    if ((retval = this->read_line(offset_inout, *lv))) {
+        sbr.share(this->lb_share_manager, lv->lv_start, lv->lv_len);
     }
 
     return retval;
