@@ -36,12 +36,26 @@
 #include <stdio.h>
 #include <fcntl.h>
 #include <ctype.h>
+#include <wordexp.h>
 
-#include <sqlite3.h>
+#include <fstream>
 
 #include "auto_fd.hh"
-#include "ptimec.hh"
 #include "lnav_util.hh"
+#include "pcrepp.hh"
+#include "lnav_config.hh"
+
+using namespace std;
+
+bool is_url(const char *fn)
+{
+    static pcrepp url_re("^(file|https?|ftps?||scp|sftp):");
+
+    pcre_context_static<30> pc;
+    pcre_input pi(fn);
+
+    return url_re.match(pc, pi);
+}
 
 std::string hash_string(const std::string &str)
 {
@@ -55,13 +69,81 @@ std::string hash_string(const std::string &str)
     return hash.to_string();
 }
 
-std::string time_ago(time_t last_time)
+std::string hash_bytes(const char *str1, size_t s1len, ...)
+{
+    byte_array<2, uint64> hash;
+    SpookyHash context;
+    va_list args;
+
+    va_start(args, s1len);
+
+    context.Init(0, 0);
+    while (str1 != NULL) {
+        context.Update(str1, s1len);
+
+        str1 = va_arg(args, const char *);
+        if (str1 == NULL) {
+            break;
+        }
+        s1len = va_arg(args, size_t);
+    }
+    context.Final(hash.out(0), hash.out(1));
+
+    va_end(args);
+
+    return hash.to_string();
+}
+
+size_t unquote(char *dst, const char *str, size_t len)
+{
+    if (str[0] == 'r' || str[0] == 'u') {
+        str += 1;
+        len -= 1;
+    }
+    char quote_char = str[0];
+    size_t index = 0;
+
+    require(str[0] == '\'' || str[0] == '"');
+
+    for (size_t lpc = 1; lpc < (len - 1); lpc++, index++) {
+        dst[index] = str[lpc];
+        if (str[lpc] == quote_char) {
+            lpc += 1;
+        }
+        else if (str[lpc] == '\\' && (lpc + 1) < len) {
+            switch (str[lpc] + 1) {
+                case 'n':
+                    dst[index] = '\n';
+                    break;
+                case 'r':
+                    dst[index] = '\r';
+                    break;
+                case 't':
+                    dst[index] = '\t';
+                    break;
+                default:
+                    dst[index] = str[lpc + 1];
+                    break;
+            }
+            lpc += 1;
+        }
+    }
+    dst[index] = '\0';
+
+    return index;
+}
+
+std::string time_ago(time_t last_time, bool convert_local)
 {
     time_t      delta, current_time = time(NULL);
     const char *fmt;
     char        buffer[64];
     int         amount;
 
+    if (convert_local) {
+        current_time = convert_log_time_to_local(current_time);
+    }
+
     delta = current_time - last_time;
     if (delta < 0) {
         return "in the future";
@@ -99,10 +181,47 @@ std::string time_ago(time_t last_time)
     return std::string(buffer);
 }
 
-/* XXX figure out how to do this with the template */
-void sqlite_close_wrapper(void *mem)
+std::string precise_time_ago(const struct timeval &tv, bool convert_local)
 {
-    sqlite3_close((sqlite3 *)mem);
+    struct timeval now, diff;
+
+    gettimeofday(&now, NULL);
+    if (convert_local) {
+        now.tv_sec = convert_log_time_to_local(now.tv_sec);
+    }
+
+    timersub(&now, &tv, &diff);
+    if (diff.tv_sec < 0) {
+        return time_ago(tv.tv_sec);
+    }
+    else if (diff.tv_sec <= 1) {
+        return "a second ago";
+    }
+    else if (diff.tv_sec < (10 * 60)) {
+        char buf[64];
+
+        if (diff.tv_sec < 60) {
+            snprintf(buf, sizeof(buf),
+                     "%2ld seconds ago",
+                     diff.tv_sec);
+        }
+        else {
+            time_t seconds = diff.tv_sec % 60;
+            time_t minutes = diff.tv_sec / 60;
+
+            snprintf(buf, sizeof(buf),
+                     "%2ld minute%s and %2ld second%s ago",
+                     minutes,
+                     minutes > 1 ? "s" : "",
+                     seconds,
+                     seconds == 1 ? "" : "s");
+        }
+
+        return string(buf);
+    }
+    else {
+        return time_ago(tv.tv_sec, convert_local);
+    }
 }
 
 std::string get_current_dir(void)
@@ -144,6 +263,37 @@ bool change_to_parent_dir(void)
     return retval;
 }
 
+void split_ws(const std::string &str, std::vector<std::string> &toks_out)
+{
+    std::stringstream ss(str);
+    std::string buf;
+
+    while (ss >> buf) {
+        toks_out.push_back(buf);
+    }
+}
+
+std::pair<std::string, std::string> split_path(const char *path, ssize_t len)
+{
+    ssize_t dir_len = len;
+
+    while (dir_len >= 0 && (path[dir_len] == '/' || path[dir_len] == '\\')) {
+        dir_len -= 1;
+    }
+
+    while (dir_len >= 0) {
+        if (path[dir_len] == '/' || path[dir_len] == '\\') {
+            return make_pair(string(path, dir_len),
+                             string(&path[dir_len + 1], len - dir_len));
+        }
+
+        dir_len -= 1;
+    }
+
+    return make_pair(path[0] == '/' ? "/" : ".",
+                     path[0] == '/' ? string(&path[1], len - 1) : string(path, len));
+}
+
 file_format_t detect_file_format(const std::string &filename)
 {
     file_format_t retval = FF_UNKNOWN;
@@ -206,7 +356,92 @@ time_t tm2sec(const struct tm *t)
     }                          /* must be a valid time */
 }
 
-bool next_format(const char *fmt[], int &index, int &locked_index)
+static const int MONSPERYEAR = 12;
+static const int SECSPERMIN = 60;
+static const int SECSPERHOUR = 60 * SECSPERMIN;
+static const int SECSPERDAY = 24 * SECSPERHOUR;
+static const int YEAR_BASE = 1900;
+static const int EPOCH_WDAY = 4;
+static const int DAYSPERWEEK = 7;
+static const int EPOCH_YEAR = 1970;
+
+#define isleap(y) ((((y) % 4) == 0 && ((y) % 100) != 0) || ((y) % 400) == 0)
+
+static const int mon_lengths[2][MONSPERYEAR] = {
+        {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
+        {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
+} ;
+
+static const int year_lengths[2] = {
+        365,
+        366
+} ;
+
+struct tm *secs2tm(time_t *tim_p, struct tm *res)
+{
+    long days, rem;
+    time_t lcltime;
+    int y;
+    int yleap;
+    const int *ip;
+
+    /* base decision about std/dst time on current time */
+    lcltime = *tim_p;
+
+    days = ((long)lcltime) / SECSPERDAY;
+    rem = ((long)lcltime) % SECSPERDAY;
+    while (rem < 0)
+    {
+        rem += SECSPERDAY;
+        --days;
+    }
+
+    /* compute hour, min, and sec */
+    res->tm_hour = (int) (rem / SECSPERHOUR);
+    rem %= SECSPERHOUR;
+    res->tm_min = (int) (rem / SECSPERMIN);
+    res->tm_sec = (int) (rem % SECSPERMIN);
+
+    /* compute day of week */
+    if ((res->tm_wday = ((EPOCH_WDAY + days) % DAYSPERWEEK)) < 0)
+        res->tm_wday += DAYSPERWEEK;
+
+    /* compute year & day of year */
+    y = EPOCH_YEAR;
+    if (days >= 0)
+    {
+        for (;;)
+        {
+            yleap = isleap(y);
+            if (days < year_lengths[yleap])
+                break;
+            y++;
+            days -= year_lengths[yleap];
+        }
+    }
+    else
+    {
+        do
+        {
+            --y;
+            yleap = isleap(y);
+            days += year_lengths[yleap];
+        } while (days < 0);
+    }
+
+    res->tm_year = y - YEAR_BASE;
+    res->tm_yday = days;
+    ip = mon_lengths[yleap];
+    for (res->tm_mon = 0; days >= ip[res->tm_mon]; ++res->tm_mon)
+        days -= ip[res->tm_mon];
+    res->tm_mday = days + 1;
+
+    res->tm_isdst = 0;
+
+    return (res);
+}
+
+bool next_format(const char * const fmt[], int &index, int &locked_index)
 {
     bool retval = true;
 
@@ -255,9 +490,11 @@ const char *std_time_fmt[] = {
 };
 
 const char *date_time_scanner::scan(const char *time_dest,
-                                    const char *time_fmt[],
-                                    struct tm *tm_out,
-                                    struct timeval &tv_out)
+                                    size_t time_len,
+                                    const char * const time_fmt[],
+                                    struct exttm *tm_out,
+                                    struct timeval &tv_out,
+                                    bool convert_local)
 {
     int  curr_time_fmt = -1;
     bool found         = false;
@@ -271,23 +508,28 @@ const char *date_time_scanner::scan(const char *time_dest,
                        curr_time_fmt,
                        this->dts_fmt_lock)) {
         *tm_out = this->dts_base_tm;
+        tm_out->et_flags = 0;
         if (time_dest[0] == '+') {
+            char time_cp[time_len + 1];
             int gmt_int, off;
 
             retval = NULL;
-            if (sscanf(time_dest, "+%d%n", &gmt_int, &off) == 1) {
+            memcpy(time_cp, time_dest, time_len);
+            time_cp[time_len] = '\0';
+            if (sscanf(time_cp, "+%d%n", &gmt_int, &off) == 1) {
                 time_t gmt = gmt_int;
 
-                if (this->dts_local_time) {
-                    localtime_r(&gmt, tm_out);
+                if (convert_local && this->dts_local_time) {
+                    localtime_r(&gmt, &tm_out->et_tm);
 #ifdef HAVE_STRUCT_TM_TM_ZONE
-                    tm_out->tm_zone = NULL;
+                    tm_out->et_tm.tm_zone = NULL;
 #endif
-                    tm_out->tm_isdst = 0;
-                    gmt = tm2sec(tm_out);
+                    tm_out->et_tm.tm_isdst = 0;
+                    gmt = tm2sec(&tm_out->et_tm);
                 }
                 tv_out.tv_sec = gmt;
                 tv_out.tv_usec = 0;
+                tm_out->et_flags = ETF_DAY_SET|ETF_MONTH_SET|ETF_YEAR_SET|ETF_MACHINE_ORIENTED|ETF_EPOCH_TIME;
 
                 this->dts_fmt_lock = curr_time_fmt;
                 this->dts_fmt_len = off;
@@ -301,25 +543,24 @@ const char *date_time_scanner::scan(const char *time_dest,
             off_t off = 0;
 
 #ifdef HAVE_STRUCT_TM_TM_ZONE
-            tm_out->tm_zone = NULL;
+            if (!this->dts_keep_base_tz) {
+                tm_out->et_tm.tm_zone = NULL;
+            }
 #endif
-            if (func(tm_out, time_dest, off, strlen(time_dest))) {
+            if (func(tm_out, time_dest, off, time_len)) {
                 retval = &time_dest[off];
 
-                if (tm_out->tm_year < 70) {
-                    tm_out->tm_year = 80;
+                if (tm_out->et_tm.tm_year < 70) {
+                    tm_out->et_tm.tm_year = 80;
                 }
-                if (this->dts_local_time) {
-                    time_t gmt = tm2sec(tm_out);
+                if (convert_local &&
+                    (this->dts_local_time || tm_out->et_flags & ETF_EPOCH_TIME)) {
+                    time_t gmt = tm2sec(&tm_out->et_tm);
 
-                    localtime_r(&gmt, tm_out);
-#ifdef HAVE_STRUCT_TM_TM_ZONE
-                    tm_out->tm_zone = NULL;
-#endif
-                    tm_out->tm_isdst = 0;
+                    this->to_localtime(gmt, *tm_out);
                 }
-                tv_out.tv_sec = tm2sec(tm_out);
-                tv_out.tv_usec = 0;
+                tv_out.tv_sec = tm2sec(&tm_out->et_tm);
+                tv_out.tv_usec = tm_out->et_nsec / 1000;
 
                 this->dts_fmt_lock = curr_time_fmt;
                 this->dts_fmt_len  = retval - time_dest;
@@ -328,46 +569,34 @@ const char *date_time_scanner::scan(const char *time_dest,
                 break;
             }
         }
-        else if ((retval = strptime(time_dest,
-                                    time_fmt[curr_time_fmt],
-                                    tm_out)) != NULL) {
-            if (time_fmt[curr_time_fmt] == time_fmt_with_zone) {
-                int lpc;
-
-                for (lpc = 0; retval[lpc] && retval[lpc] != ' '; lpc++) {
+        else {
+            off_t off = 0;
 
+            if (ptime_fmt(time_fmt[curr_time_fmt], tm_out, time_dest, off, time_len)) {
+                retval = &time_dest[off];
+                if (tm_out->et_tm.tm_year < 70) {
+                    tm_out->et_tm.tm_year = 80;
                 }
-                if (retval[lpc] == ' ' &&
-                    sscanf(&retval[lpc], "%d", &tm_out->tm_year) == 1) {
-                    lpc += 1;
-                    for (; retval[lpc] && isdigit(retval[lpc]); lpc++) {
+                if (convert_local &&
+                    (this->dts_local_time || tm_out->et_flags & ETF_EPOCH_TIME)) {
+                    time_t gmt = tm2sec(&tm_out->et_tm);
 
-                    }
-                    retval = &retval[lpc];
-                }
-            }
-
-            if (tm_out->tm_year < 70) {
-                tm_out->tm_year = 80;
-            }
-            if (this->dts_local_time) {
-                time_t gmt = tm2sec(tm_out);
-
-                localtime_r(&gmt, tm_out);
+                    this->to_localtime(gmt, *tm_out);
 #ifdef HAVE_STRUCT_TM_TM_ZONE
-                tm_out->tm_zone = NULL;
+                    tm_out->et_tm.tm_zone = NULL;
 #endif
-                tm_out->tm_isdst = 0;
+                    tm_out->et_tm.tm_isdst = 0;
+                }
+
+                tv_out.tv_sec = tm2sec(&tm_out->et_tm);
+                tv_out.tv_usec = tm_out->et_nsec / 1000;
+
+                this->dts_fmt_lock = curr_time_fmt;
+                this->dts_fmt_len  = retval - time_dest;
+
+                found = true;
+                break;
             }
-
-            tv_out.tv_sec = tm2sec(tm_out);
-            tv_out.tv_usec = 0;
-
-            this->dts_fmt_lock = curr_time_fmt;
-            this->dts_fmt_len  = retval - time_dest;
-
-            found = true;
-            break;
         }
     }
 
@@ -378,19 +607,17 @@ const char *date_time_scanner::scan(const char *time_dest,
     if (retval != NULL) {
         /* Try to pull out the milli/micro-second value. */
         if (retval[0] == '.' || retval[0] == ',') {
-            int sub_seconds = 0, sub_len = 0;
+            off_t off = (retval - time_dest) + 1;
 
-            if (sscanf(retval + 1, "%d%n", &sub_seconds, &sub_len) == 1) {
-                switch (sub_len) {
-                case 3:
-                    tv_out.tv_usec = sub_seconds * 1000;
-                    this->dts_fmt_len += 1 + sub_len;
-                    break;
-                case 6:
-                    tv_out.tv_usec = sub_seconds;
-                    this->dts_fmt_len += 1 + sub_len;
-                    break;
-                }
+            if (ptime_f(tm_out, time_dest, off, time_len)) {
+                tv_out.tv_usec = tm_out->et_nsec / 1000;
+                this->dts_fmt_len += 7;
+                retval += 7;
+            }
+            else if (ptime_F(tm_out, time_dest, off, time_len)) {
+                tv_out.tv_usec = tm_out->et_nsec / 1000;
+                this->dts_fmt_len += 4;
+                retval += 4;
             }
         }
     }
@@ -427,3 +654,65 @@ size_t strtonum<long>(long &num_out, const char *string, size_t len);
 
 template
 size_t strtonum<int>(int &num_out, const char *string, size_t len);
+
+string build_path(const vector<string> &paths)
+{
+    string retval;
+
+    for (vector<string>::const_iterator path_iter = paths.begin();
+         path_iter != paths.end();
+         ++path_iter) {
+        if (path_iter->empty()) {
+            continue;
+        }
+        if (!retval.empty()) {
+            retval += ":";
+        }
+        retval += *path_iter;
+    }
+    retval += ":" + string(getenv("PATH"));
+    return retval;
+}
+
+bool read_file(const char *filename, string &out)
+{
+    std::ifstream sql_file(filename);
+
+    if (sql_file) {
+        out.assign((std::istreambuf_iterator<char>(sql_file)),
+                   std::istreambuf_iterator<char>());
+        return true;
+    }
+
+    return false;
+}
+
+bool wordexperr(int rc, string &msg)
+{
+    switch (rc) {
+        case WRDE_BADCHAR:
+            msg = "error: invalid filename character";
+            return false;
+
+        case WRDE_CMDSUB:
+            msg = "error: command substitution is not allowed";
+            return false;
+
+        case WRDE_BADVAL:
+            msg = "error: unknown environment variable in file name";
+            return false;
+
+        case WRDE_NOSPACE:
+            msg = "error: out of memory";
+            return false;
+
+        case WRDE_SYNTAX:
+            msg = "error: invalid syntax";
+            return false;
+
+        default:
+            break;
+    }
+
+    return true;
+}
