@@ -29,6 +29,7 @@
 
 #include "config.h"
 
+#include "lnav_log.hh"
 #include "sql_util.hh"
 #include "log_vtab_impl.hh"
 
@@ -38,7 +39,7 @@ using namespace std;
 
 static struct log_cursor log_cursor_latest;
 
-static sql_progress_callback_t vtab_progress_callback;
+sql_progress_callback_t log_vtab_progress_callback;
 
 static const char *type_to_string(int type)
 {
@@ -53,32 +54,34 @@ static const char *type_to_string(int type)
         return "text";
     }
 
-    assert("Invalid sqlite type");
+    ensure("Invalid sqlite type");
 
     return NULL;
 }
 
-static string declare_table_statement(log_vtab_impl *vi)
+std::string log_vtab_impl::get_table_statement(void)
 {
     std::vector<log_vtab_impl::vtab_column> cols;
     std::vector<log_vtab_impl::vtab_column>::const_iterator iter;
     std::ostringstream oss;
 
-    oss << "CREATE TABLE unused (\n"
+    oss << "CREATE TABLE " << this->get_name().to_string() << " (\n"
         << "  log_line integer PRIMARY KEY,\n"
         << "  log_part text collate naturalnocase,\n"
         << "  log_time datetime,\n"
+        << "  log_actual_time datetime hidden,\n"
         << "  log_idle_msecs int,\n"
-        << "  log_level text,\n";
-    vi->get_columns(cols);
-    vi->vi_column_count = cols.size();
+        << "  log_level text collate loglevel,\n"
+        << "  log_mark boolean,\n";
+    this->get_columns(cols);
+    this->vi_column_count = cols.size();
     for (iter = cols.begin(); iter != cols.end(); iter++) {
         auto_mem<char, sqlite3_free> coldecl;
+        auto_mem<char, sqlite3_free> colname;
 
-        assert(iter->vc_name != NULL);
-
-        coldecl = sqlite3_mprintf("  %Q %s %s collate %Q,\n",
-                                  iter->vc_name,
+        colname = sql_quote_ident(iter->vc_name.c_str());
+        coldecl = sqlite3_mprintf("  %s %s %s collate %Q,\n",
+                                  colname.in(),
                                   type_to_string(iter->vc_type),
                                   iter->vc_hidden ? "hidden" : "",
                                   (iter->vc_collator == NULL ||
@@ -87,8 +90,9 @@ static string declare_table_statement(log_vtab_impl *vi)
         oss << coldecl;
     }
     oss << "  log_path text hidden collate naturalnocase,\n"
-        << "  log_text text hidden\n"
-        << ");";
+        << "  log_text text hidden,\n"
+        << "  log_body text hidden\n"
+        << ");\n";
 
     return oss.str();
 }
@@ -104,6 +108,7 @@ struct vtab {
 struct vtab_cursor {
     sqlite3_vtab_cursor        base;
     struct log_cursor          log_cursor;
+    shared_buffer_ref          log_msg;
     std::vector<logline_value> line_values;
 };
 
@@ -130,19 +135,20 @@ static int vt_create(sqlite3 *db,
     p_vt->db = db;
 
     /* Declare the vtable's structure */
-    p_vt->vi = vm->lookup_impl(argv[3]);
+    p_vt->vi = vm->lookup_impl(intern_string::lookup(argv[3]));
     if (p_vt->vi == NULL) {
         return SQLITE_ERROR;
     }
     p_vt->tc = vm->get_view();
     p_vt->lss = vm->get_source();
-    rc        = sqlite3_declare_vtab(db, declare_table_statement(
-                                         p_vt->vi).c_str());
+    rc        = sqlite3_declare_vtab(db, p_vt->vi->get_table_statement().c_str());
 
     /* Success. Set *pp_vt and return */
     *pp_vt = &p_vt->base;
 
-    return SQLITE_OK;
+    log_debug("creating log format table: %s = %p", argv[3], p_vt);
+
+    return rc;
 }
 
 static int vt_destructor(sqlite3_vtab *p_svt)
@@ -180,16 +186,17 @@ static int vt_open(sqlite3_vtab *p_svt, sqlite3_vtab_cursor **pp_cursor)
 
     p_vt->base.zErrMsg = NULL;
 
-    vtab_cursor *p_cur = (vtab_cursor *)new vtab_cursor();
+    vtab_cursor *p_cur = new vtab_cursor();
 
     *pp_cursor = (sqlite3_vtab_cursor *)p_cur;
 
     p_cur->base.pVtab = p_svt;
     p_cur->log_cursor.lc_curr_line = vis_line_t(-1);
+    p_cur->log_cursor.lc_end_line = vis_line_t(p_vt->lss->text_line_count());
     p_cur->log_cursor.lc_sub_index = 0;
     vt_next((sqlite3_vtab_cursor *)p_cur);
 
-    return p_cur ? SQLITE_OK : SQLITE_NOMEM;
+    return SQLITE_OK;
 }
 
 static int vt_close(sqlite3_vtab_cursor *cur)
@@ -205,9 +212,8 @@ static int vt_close(sqlite3_vtab_cursor *cur)
 static int vt_eof(sqlite3_vtab_cursor *cur)
 {
     vtab_cursor *vc = (vtab_cursor *)cur;
-    vtab *       vt = (vtab *)cur->pVtab;
 
-    return vc->log_cursor.lc_curr_line == (int)vt->lss->text_line_count();
+    return vc->log_cursor.is_eof();
 }
 
 static int vt_next(sqlite3_vtab_cursor *cur)
@@ -220,8 +226,7 @@ static int vt_next(sqlite3_vtab_cursor *cur)
     do {
         log_cursor_latest = vc->log_cursor;
         if (((log_cursor_latest.lc_curr_line % 1024) == 0) &&
-            vtab_progress_callback(log_cursor_latest)) {
-            done = true;
+            log_vtab_progress_callback(log_cursor_latest)) {
             break;
         }
         done = vt->vi->next(vc->log_cursor, *vt->lss);
@@ -239,7 +244,7 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
     logfile *         lf = vt->lss->find(cl);
     logfile::iterator ll = lf->begin() + cl;
 
-    assert(col >= 0);
+    require(col >= 0);
 
     /* Just return the ordinal of the column requested. */
     switch (col) {
@@ -252,24 +257,20 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
     case VT_COL_PARTITION:
     {
         vis_bookmarks &vb = vt->tc->get_bookmarks();
-        bookmark_vector<vis_line_t> &bv = vb[&textview_curses::BM_USER];
+        bookmark_vector<vis_line_t> &bv = vb[&textview_curses::BM_PARTITION];
         bookmark_vector<vis_line_t>::iterator iter;
-        vis_line_t prev_line;
-        char part_name[64];
-        int index;
+        vis_line_t curr_line;
 
-        prev_line = vis_line_t(vc->log_cursor.lc_curr_line);
-        ++prev_line;
-        iter = lower_bound(bv.begin(), bv.end(), prev_line);
-        index = distance(bv.begin(), iter);
-        snprintf(part_name, sizeof(part_name), "p.%d", index);
+        curr_line = vis_line_t(vc->log_cursor.lc_curr_line);
+        iter = lower_bound(bv.begin(), bv.end(), curr_line);
 
-        if (iter == bv.begin()) {
-            sqlite3_result_text(ctx, part_name, strlen(part_name), SQLITE_TRANSIENT);
+        if (bv.empty() || (iter != bv.end() && curr_line < *iter)) {
+            sqlite3_result_null(ctx);
         }
         else {
-            --iter;
-
+            if (iter == bv.end() || *iter != curr_line) {
+                --iter;
+            }
             content_line_t part_line = vt->lss->at(*iter);
             std::map<content_line_t, bookmark_metadata> &bm_meta = vt->lss->get_user_bookmark_metadata();
             std::map<content_line_t, bookmark_metadata>::iterator meta_iter;
@@ -282,7 +283,7 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
                                     SQLITE_TRANSIENT);
             }
             else {
-                sqlite3_result_text(ctx, part_name, strlen(part_name), SQLITE_TRANSIENT);
+                sqlite3_result_null(ctx);
             }
         }
     }
@@ -297,6 +298,39 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
     }
     break;
 
+        case VT_COL_LOG_ACTUAL_TIME: {
+            char buffer[64];
+
+            if (ll->is_time_skewed()) {
+                if (vc->line_values.empty()) {
+                    lf->read_full_message(ll, vc->log_msg);
+                    vt->vi->extract(lf, vc->log_msg, vc->line_values);
+                }
+
+                struct line_range time_range;
+
+                time_range = find_string_attr_range(
+                    vt->vi->vi_attrs, &logline::L_TIMESTAMP);
+
+                const char *time_src = vc->log_msg.get_data() + time_range.lr_start;
+                struct timeval actual_tv;
+                struct exttm tm;
+
+                if (lf->get_format()->lf_date_time.scan(
+                    time_src, time_range.length(),
+                    lf->get_format()->get_timestamp_formats(),
+                    &tm, actual_tv,
+                    false)) {
+                    sql_strftime(buffer, sizeof(buffer), actual_tv);
+                }
+            }
+            else {
+                sql_strftime(buffer, sizeof(buffer), ll->get_time(), ll->get_millis());
+            }
+            sqlite3_result_text(ctx, buffer, strlen(buffer), SQLITE_TRANSIENT);
+            break;
+        }
+
     case VT_COL_IDLE_MSECS:
         if (vc->log_cursor.lc_curr_line == 0) {
             sqlite3_result_int64(ctx, 0);
@@ -313,7 +347,7 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
             prev_time      += prev_ll->get_millis();
             curr_line_time  = ll->get_time() * 1000ULL;
             curr_line_time += ll->get_millis();
-            assert(curr_line_time >= prev_time);
+            // require(curr_line_time >= prev_time);
             sqlite3_result_int64(ctx, curr_line_time - prev_time);
         }
         break;
@@ -329,38 +363,73 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
     }
     break;
 
+    case VT_COL_MARK:
+    {
+        sqlite3_result_int(ctx, ll->is_marked());
+    }
+    break;
+
     default:
         if (col > (VT_COL_MAX + vt->vi->vi_column_count - 1)) {
             int post_col_number = col -
                                   (VT_COL_MAX + vt->vi->vi_column_count -
                                    1) - 1;
 
-            if (post_col_number == 0) {
-                const string &fn = lf->get_filename();
+            switch (post_col_number) {
+                case 0: {
+                    const string &fn = lf->get_filename();
 
-                sqlite3_result_text(ctx,
-                                    fn.c_str(),
-                                    fn.length(),
-                                    SQLITE_STATIC);
-            }
-            else {
-                string line;
+                    sqlite3_result_text(ctx,
+                                        fn.c_str(),
+                                        fn.length(),
+                                        SQLITE_STATIC);
+                    break;
+                }
+                case 1: {
+                    shared_buffer_ref line;
 
-                lf->read_line(ll, line);
-                sqlite3_result_text(ctx,
-                                    line.c_str(),
-                                    line.length(),
-                                    SQLITE_TRANSIENT);
+                    lf->read_full_message(ll, line);
+                    sqlite3_result_text(ctx,
+                                        line.get_data(),
+                                        line.length(),
+                                        SQLITE_TRANSIENT);
+                    break;
+                }
+                case 2: {
+                    if (vc->line_values.empty()) {
+                        logfile::iterator line_iter;
+
+                        line_iter = lf->begin() + cl;
+                        lf->read_full_message(line_iter, vc->log_msg);
+                        vt->vi->extract(lf, vc->log_msg, vc->line_values);
+                    }
+
+                    struct line_range body_range;
+
+                    body_range = find_string_attr_range(
+                        vt->vi->vi_attrs, &textview_curses::SA_BODY);
+                    if (!body_range.is_valid()) {
+                        sqlite3_result_null(ctx);
+                    }
+                    else {
+                        const char *msg_start = vc->log_msg.get_data();
+
+                        sqlite3_result_text(ctx,
+                                            &msg_start[body_range.lr_start],
+                                            body_range.length(),
+                                            SQLITE_TRANSIENT);
+                    }
+                    break;
+                }
             }
         }
         else {
             if (vc->line_values.empty()) {
                 logfile::iterator line_iter;
-                string            line, value;
 
                 line_iter = lf->begin() + cl;
-                lf->read_line(line_iter, line);
-                vt->vi->extract(lf, line, vc->line_values);
+                lf->read_full_message(line_iter, vc->log_msg);
+                vt->vi->extract(lf, vc->log_msg, vc->line_values);
             }
 
             size_t sub_col = col - VT_COL_MAX;
@@ -374,24 +443,57 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
                 case logline_value::VALUE_NULL:
                     sqlite3_result_null(ctx);
                     break;
-                case logline_value::VALUE_TEXT:
+                case logline_value::VALUE_JSON:
+                case logline_value::VALUE_TEXT: {
                     sqlite3_result_text(ctx,
-                                        lv_iter->lv_string.c_str(),
-                                        lv_iter->lv_string.length(),
+                                        lv_iter->text_value(),
+                                        lv_iter->text_length(),
                                         SQLITE_TRANSIENT);
                     break;
+                }
+                case logline_value::VALUE_QUOTED:
+                    if (lv_iter->lv_sbr.length() == 0) {
+                        sqlite3_result_text(ctx, "", 0, SQLITE_STATIC);
+                    }
+                    else {
+                        const char *text_value = lv_iter->lv_sbr.get_data();
+                        size_t text_len = lv_iter->lv_sbr.length();
+
+                        switch (text_value[0]) {
+                        case '\'':
+                        case '"': {
+                            char *val = (char *)sqlite3_malloc(text_len);
+
+                            if (val == NULL) {
+                                sqlite3_result_error_nomem(ctx);
+                            }
+                            else {
+                                size_t unquoted_len = unquote(val, text_value, text_len);
+                                sqlite3_result_text(ctx, val, unquoted_len, sqlite3_free);
+                            }
+                            break;
+                        }
+                        default: {
+                            sqlite3_result_text(ctx, text_value,
+                                lv_iter->lv_sbr.length(), SQLITE_TRANSIENT);
+                            break;
+                        }
+                        }
+                    }
+                    break;
 
                 case logline_value::VALUE_BOOLEAN:
                 case logline_value::VALUE_INTEGER:
-                    sqlite3_result_int64(ctx, lv_iter->lv_number.i);
+                    sqlite3_result_int64(ctx, lv_iter->lv_value.i);
                     break;
 
                 case logline_value::VALUE_FLOAT:
-                    sqlite3_result_double(ctx, lv_iter->lv_number.d);
+                    sqlite3_result_double(ctx, lv_iter->lv_value.d);
                     break;
 
                 case logline_value::VALUE_UNKNOWN:
-                    assert(0);
+                case logline_value::VALUE__MAX:
+                    require(0);
                     break;
                 }
             }
@@ -415,18 +517,205 @@ static int vt_rowid(sqlite3_vtab_cursor *cur, sqlite_int64 *p_rowid)
     return SQLITE_OK;
 }
 
+void log_cursor::update(unsigned char op, vis_line_t vl, bool exact)
+{
+    if (vl < 0) {
+        vl = vis_line_t(-1);
+    }
+    switch (op) {
+    case SQLITE_INDEX_CONSTRAINT_EQ:
+        if (vl < this->lc_end_line) {
+            this->lc_curr_line = vl;
+            this->lc_end_line = vis_line_t(this->lc_curr_line + 1);
+        }
+        break;
+    case SQLITE_INDEX_CONSTRAINT_GE:
+        this->lc_curr_line = vl;
+        break;
+    case SQLITE_INDEX_CONSTRAINT_GT:
+        this->lc_curr_line = vis_line_t(vl + (exact ? 1 : 0));
+        break;
+    case SQLITE_INDEX_CONSTRAINT_LE:
+        this->lc_end_line = vis_line_t(vl + (exact ? 1 : 0));
+        break;
+    case SQLITE_INDEX_CONSTRAINT_LT:
+        this->lc_end_line = vl;
+        break;
+    }
+}
+
 static int vt_filter(sqlite3_vtab_cursor *p_vtc,
                      int idxNum, const char *idxStr,
                      int argc, sqlite3_value **argv)
 {
+    vtab_cursor *p_cur = (vtab_cursor *)p_vtc;
+    vtab *       vt = (vtab *)p_vtc->pVtab;
+    sqlite3_index_info::sqlite3_index_constraint *index = (
+        sqlite3_index_info::sqlite3_index_constraint *)idxStr;
+
+    log_info("(%p) filter called: %d", vt, idxNum);
+    p_cur->log_cursor.lc_curr_line = vis_line_t(-1);
+    p_cur->log_cursor.lc_end_line = vis_line_t(vt->lss->text_line_count());
+    vt_next(p_vtc);
+
+    if (!idxNum) {
+        return SQLITE_OK;
+    }
+
+    for (int lpc = 0; lpc < idxNum; lpc++) {
+        switch (index[lpc].iColumn) {
+        case VT_COL_LINE_NUMBER:
+            p_cur->log_cursor.update(index[lpc].op,
+                vis_line_t(sqlite3_value_int64(argv[lpc])));
+            break;
+
+        case VT_COL_LOG_TIME: {
+            const unsigned char *datestr = sqlite3_value_text(argv[lpc]);
+            date_time_scanner dts;
+            struct timeval tv;
+            struct exttm mytm;
+            vis_line_t vl;
+
+            dts.scan((const char *)datestr, strlen((const char *)datestr), NULL, &mytm, tv);
+            if ((vl = vt->lss->find_from_time(tv)) == -1) {
+                p_cur->log_cursor.lc_curr_line = p_cur->log_cursor.lc_end_line;
+            }
+            else {
+                p_cur->log_cursor.update(index[lpc].op, vl, false);
+            }
+            break;
+        }
+
+        }
+    }
+
     return SQLITE_OK;
 }
 
 static int vt_best_index(sqlite3_vtab *tab, sqlite3_index_info *p_info)
 {
+    std::vector<sqlite3_index_info::sqlite3_index_constraint> indexes;
+    int argvInUse = 0;
+    vtab *vt = (vtab *) tab;
+
+    log_info("(%p) best index called: nConstraint=%d", tab, p_info->nConstraint);
+    if (!vt->vi->vi_supports_indexes) {
+        return SQLITE_OK;
+    }
+    for (int lpc = 0; lpc < p_info->nConstraint; lpc++) {
+        if (!p_info->aConstraint[lpc].usable ||
+            p_info->aConstraint[lpc].op == SQLITE_INDEX_CONSTRAINT_MATCH) {
+            continue;
+        }
+
+        switch (p_info->aConstraint[lpc].iColumn) {
+        case VT_COL_LINE_NUMBER:
+            argvInUse += 1;
+            indexes.push_back(p_info->aConstraint[lpc]);
+            p_info->aConstraintUsage[lpc].argvIndex = argvInUse;
+            break;
+        }
+    }
+
+    if (!argvInUse) {
+        for (int lpc = 0; lpc < p_info->nConstraint; lpc++) {
+            if (!p_info->aConstraint[lpc].usable ||
+                p_info->aConstraint[lpc].op == SQLITE_INDEX_CONSTRAINT_MATCH) {
+                continue;
+            }
+
+            switch (p_info->aConstraint[lpc].iColumn) {
+            case VT_COL_LOG_TIME:
+                argvInUse += 1;
+                indexes.push_back(p_info->aConstraint[lpc]);
+                p_info->aConstraintUsage[lpc].argvIndex = argvInUse;
+                break;
+            }
+        }
+    }
+
+    if (argvInUse) {
+        sqlite3_index_info::sqlite3_index_constraint *index_copy;
+        size_t len = indexes.size() * sizeof(*index_copy);
+
+        log_info("found index, passing %d args", argvInUse);
+
+        index_copy = (sqlite3_index_info::sqlite3_index_constraint *)
+            sqlite3_malloc(len);
+        if (!index_copy) {
+            return SQLITE_NOMEM;
+        }
+        memcpy(index_copy, &indexes[0], len);
+        p_info->idxNum = argvInUse;
+        p_info->idxStr = (char *) index_copy;
+        p_info->needToFreeIdxStr = 1;
+        p_info->estimatedCost = 1.0;
+    }
+
     return SQLITE_OK;
 }
 
+static int vt_update(sqlite3_vtab *tab,
+                     int argc,
+                     sqlite3_value **argv,
+                     sqlite_int64 *rowid_out)
+{
+    vtab *vt = (vtab *)tab;
+    int retval = SQLITE_READONLY;
+
+    if (argc > 1 && sqlite3_value_type(argv[0]) != SQLITE_NULL &&
+        sqlite3_value_int64(argv[0]) == sqlite3_value_int64(argv[1])) {
+        int64_t rowid = sqlite3_value_int64(argv[0]) >> 8;
+        int val = sqlite3_value_int(argv[2 + VT_COL_MARK]);
+        vis_line_t vrowid(rowid);
+
+        std::map<content_line_t, bookmark_metadata> &bm = vt->lss->get_user_bookmark_metadata();
+        const unsigned char *part_name = sqlite3_value_text(argv[2 + VT_COL_PARTITION]);
+
+        bookmark_vector<vis_line_t> &bv = vt->tc->get_bookmarks()[
+                &textview_curses::BM_PARTITION];
+        bookmark_vector<vis_line_t>::iterator part_iter;
+        bool set_name = false;
+
+        if ((part_iter = find(bv.begin(), bv.end(), vrowid)) != bv.end()) {
+            if (part_name == NULL) {
+                vt->tc->set_user_mark(&textview_curses::BM_PARTITION, vrowid, false);
+                bm.erase(vt->lss->at(vrowid));
+            }
+            else {
+                set_name = true;
+            }
+        }
+        else if (part_name != NULL) {
+            vt->tc->set_user_mark(&textview_curses::BM_PARTITION, vrowid, true);
+            set_name = true;
+        }
+
+        if (set_name) {
+            bookmark_metadata &line_meta = bm[vt->lss->at(vrowid)];
+
+            line_meta.bm_name = string((const char *) part_name);
+        }
+
+        vt->tc->set_user_mark(&textview_curses::BM_USER, vis_line_t(rowid), val);
+        rowid += 1;
+        while ((size_t)rowid < vt->lss->text_line_count()) {
+            vis_line_t vl(rowid);
+            content_line_t cl = vt->lss->at(vl);
+            logline *ll = vt->lss->find_line(cl);
+            if (!ll->is_continued()) {
+                break;
+            }
+            vt->tc->set_user_mark(&textview_curses::BM_USER, vl, val);
+            rowid += 1;
+        }
+
+        retval = SQLITE_OK;
+    }
+
+    return retval;
+}
+
 static sqlite3_module vtab_module = {
     0,              /* iVersion */
     vt_create,      /* xCreate       - create a vtable */
@@ -441,7 +730,7 @@ static sqlite3_module vtab_module = {
     vt_eof,         /* xEof          - inidicate end of result set*/
     vt_column,      /* xColumn       - read data */
     vt_rowid,       /* xRowid        - read data */
-    NULL,           /* xUpdate       - write data */
+    vt_update,      /* xUpdate       - write data */
     NULL,           /* xBegin        - begin transaction */
     NULL,           /* xSync         - sync transaction */
     NULL,           /* xCommit       - commit transaction */
@@ -453,8 +742,8 @@ static int progress_callback(void *ptr)
 {
     int retval = 0;
 
-    if (vtab_progress_callback != NULL) {
-        retval = vtab_progress_callback(log_cursor_latest);
+    if (log_vtab_progress_callback != NULL) {
+        retval = log_vtab_progress_callback(log_cursor_latest);
     }
 
     return retval;
@@ -462,12 +751,10 @@ static int progress_callback(void *ptr)
 
 log_vtab_manager::log_vtab_manager(sqlite3 *memdb,
                                    textview_curses &tc,
-                                   logfile_sub_source &lss,
-                                   sql_progress_callback_t pc)
+                                   logfile_sub_source &lss)
     : vm_db(memdb), vm_textview(tc), vm_source(lss)
 {
     sqlite3_create_module(this->vm_db, "log_vtab_impl", &vtab_module, this);
-    vtab_progress_callback = pc;
     sqlite3_progress_handler(memdb, 32, progress_callback, NULL);
 }
 
@@ -477,15 +764,15 @@ string log_vtab_manager::register_vtab(log_vtab_impl *vi)
 
     if (this->vm_impls.find(vi->get_name()) == this->vm_impls.end()) {
         auto_mem<char> errmsg(sqlite3_free);
-        char *         sql;
+        auto_mem<char> sql(sqlite3_free);
         int            rc;
 
         this->vm_impls[vi->get_name()] = vi;
 
         sql = sqlite3_mprintf("CREATE VIRTUAL TABLE %s "
                               "USING log_vtab_impl(%s)",
-                              vi->get_name().c_str(),
-                              vi->get_name().c_str());
+                              vi->get_name().get(),
+                              vi->get_name().get());
         rc = sqlite3_exec(this->vm_db,
                           sql,
                           NULL,
@@ -494,8 +781,6 @@ string log_vtab_manager::register_vtab(log_vtab_impl *vi)
         if (rc != SQLITE_OK) {
             retval = errmsg;
         }
-
-        sqlite3_free(sql);
     }
     else {
         retval = "a table with the given name already exists";
@@ -504,26 +789,23 @@ string log_vtab_manager::register_vtab(log_vtab_impl *vi)
     return retval;
 }
 
-string log_vtab_manager::unregister_vtab(std::string name)
+string log_vtab_manager::unregister_vtab(intern_string_t name)
 {
     string retval = "";
 
     if (this->vm_impls.find(name) == this->vm_impls.end()) {
-        retval = "unknown log line table -- " + name;
+        retval = "unknown log line table -- " + name.to_string();
     }
     else {
-        char *sql;
+        auto_mem<char> sql(sqlite3_free);
         int   rc;
 
-        sql = sqlite3_mprintf("DROP TABLE %s ", name.c_str());
+        sql = sqlite3_mprintf("DROP TABLE %s ", name.get());
         rc  = sqlite3_exec(this->vm_db,
                            sql,
                            NULL,
                            NULL,
                            NULL);
-        assert(rc == SQLITE_OK);
-
-        sqlite3_free(sql);
 
         this->vm_impls.erase(name);
     }
