@@ -41,6 +41,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/param.h>
+#include <sys/resource.h>
 
 #include <time.h>
 
@@ -50,64 +51,71 @@
 using namespace std;
 
 static const size_t MAX_UNRECOGNIZED_LINES = 1000;
+static const size_t INDEX_RESERVE_INCREMENT = 1024;
 
-logfile::logfile(string filename, auto_fd fd)
+logfile::logfile(const string &filename, logfile_open_options &loo)
 throw (error)
     : lf_filename(filename),
       lf_index_time(0),
       lf_index_size(0),
-      lf_is_closed(false)
+      lf_is_closed(false),
+      lf_logline_observer(NULL),
+      lf_logfile_observer(NULL),
+      lf_longest_line(0)
 {
-    int reserve_size = 100;
-
     require(filename.size() > 0);
 
     this->lf_time_offset.tv_sec = 0;
     this->lf_time_offset.tv_usec = 0;
     
     memset(&this->lf_stat, 0, sizeof(this->lf_stat));
-    if (fd == -1) {
+    if (loo.loo_fd == -1) {
         char resolved_path[PATH_MAX];
 
         errno = 0;
         if (realpath(filename.c_str(), resolved_path) == NULL) {
             throw error(resolved_path, errno);
         }
-        filename = resolved_path;
 
-        if (stat(filename.c_str(), &this->lf_stat) == -1) {
+        if (stat(resolved_path, &this->lf_stat) == -1) {
             throw error(filename, errno);
         }
-        reserve_size = this->lf_stat.st_size / 100;
 
         if (!S_ISREG(this->lf_stat.st_mode)) {
             throw error(filename, EINVAL);
         }
 
-        if ((fd = open(filename.c_str(), O_RDONLY)) == -1) {
+        if ((loo.loo_fd = open(resolved_path, O_RDONLY)) == -1) {
             throw error(filename, errno);
         }
 
-        if (fcntl(fd, F_SETFD, FD_CLOEXEC) == -1) {
-            throw error(filename, errno);
-        }
+        loo.loo_fd.close_on_exec();
+
+        log_info("Creating logfile: fd=%d; size=%d; mtime=%d; filename=%s",
+                 (int) loo.loo_fd,
+                 this->lf_stat.st_size,
+                 this->lf_stat.st_mtime,
+                 filename.c_str());
 
         this->lf_valid_filename = true;
     }
     else {
-        fstat(fd, &this->lf_stat);
+        log_perror(fstat(loo.loo_fd, &this->lf_stat));
         this->lf_valid_filename = false;
     }
 
     this->lf_content_id = hash_string(this->lf_filename);
-    this->lf_line_buffer.set_fd(fd);
-    this->lf_index.reserve(reserve_size);
+    this->lf_line_buffer.set_fd(loo.loo_fd);
+    this->lf_index.reserve(INDEX_RESERVE_INCREMENT);
+
+    this->lf_options = loo;
 
     ensure(this->invariant());
 }
 
 logfile::~logfile()
-{ }
+{
+}
 
 bool logfile::exists(void) const
 {
@@ -122,18 +130,31 @@ bool logfile::exists(void) const
     }
 
     return this->lf_stat.st_dev == st.st_dev &&
-           this->lf_stat.st_ino == st.st_ino;
+           this->lf_stat.st_ino == st.st_ino &&
+           this->lf_stat.st_size <= st.st_size;
 }
 
-void logfile::process_prefix(off_t offset, char *prefix, int len)
+void logfile::set_format_base_time(log_format *lf)
 {
-    bool found = false;
+    time_t file_time = this->lf_line_buffer.get_file_time();
+
+    if (file_time == 0) {
+        file_time = this->lf_stat.st_mtime;
+    }
+    lf->lf_date_time.set_base_time(file_time);
+}
+
+void logfile::process_prefix(off_t offset, shared_buffer_ref &sbr)
+{
+    log_format::scan_result_t found = log_format::SCAN_NO_MATCH;
+    size_t prescan_size = this->lf_index.size();
 
     if (this->lf_format.get() != NULL) {
         /* We've locked onto a format, just use that scanner. */
-        found = this->lf_format->scan(this->lf_index, offset, prefix, len);
+        found = this->lf_format->scan(this->lf_index, offset, sbr);
     }
-    else if (this->lf_index.size() < MAX_UNRECOGNIZED_LINES) {
+    else if (this->lf_options.loo_detect_format &&
+             this->lf_index.size() < MAX_UNRECOGNIZED_LINES) {
         vector<log_format *> &root_formats =
             log_format::get_root_formats();
         vector<log_format *>::iterator iter;
@@ -143,27 +164,30 @@ void logfile::process_prefix(off_t offset, char *prefix, int len)
          * are sufficiently different that there are no ambiguities...
          */
         for (iter = root_formats.begin();
-             iter != root_formats.end() && !found;
+             iter != root_formats.end() && (found != log_format::SCAN_MATCH);
              ++iter) {
-            if (!(*iter)->match_name(this->lf_filename))
+            if (!(*iter)->match_name(this->lf_filename)) {
                 continue;
+            }
 
             (*iter)->clear();
-            (*iter)->lf_date_time.set_base_time(this->lf_line_buffer.get_file_time());
-            if ((*iter)->lf_date_time.dts_base_time == 0) {
-                (*iter)->lf_date_time.set_base_time(this->lf_stat.st_mtime);
-            }
-            if ((*iter)->scan(this->lf_index, offset, prefix, len)) {
+            this->set_format_base_time(*iter);
+            found = (*iter)->scan(this->lf_index, offset, sbr);
+            if (found == log_format::SCAN_MATCH) {
 #if 0
                 require(this->lf_index.size() == 1 ||
                        (this->lf_index[this->lf_index.size() - 2] <
                         this->lf_index[this->lf_index.size() - 1]));
 #endif
+                log_info("%s:%d:log format found -- %s",
+                    this->lf_filename.c_str(),
+                    this->lf_index.size(),
+                    (*iter)->get_name().get());
 
                 this->lf_format =
                     auto_ptr<log_format>((*iter)->specialized());
-                this->lf_content_id = hash_string(string(prefix, len));
-                found = true;
+                this->set_format_base_time(this->lf_format.get());
+                this->lf_content_id = hash_string(string(sbr.get_data(), sbr.length()));
 
                 /*
                  * We'll go ahead and assume that any previous lines were
@@ -180,59 +204,100 @@ void logfile::process_prefix(off_t offset, char *prefix, int len)
         }
     }
 
-    if (found) {
-        if (this->lf_index.size() >= 2) {
-            logline &second_to_last = this->lf_index[this->lf_index.size() - 2];
-            logline &latest = this->lf_index.back();
+    switch (found) {
+        case log_format::SCAN_MATCH:
+            if (prescan_size > 0 && prescan_size < this->lf_index.size()) {
+                logline &second_to_last = this->lf_index[prescan_size - 1];
+                logline &latest = this->lf_index[prescan_size];
 
-            if (latest < second_to_last) {
-                latest.set_time(second_to_last.get_time());
-                latest.set_millis(second_to_last.get_millis());
+                if (latest < second_to_last) {
+                    log_debug("%s:%d: out-of-time-order line detected %d.%03d < %d.%03d",
+                              this->lf_filename.c_str(),
+                              prescan_size,
+                              latest.get_time(),
+                              latest.get_millis(),
+                              second_to_last.get_time(),
+                              second_to_last.get_millis());
+                    for (size_t lpc = prescan_size; lpc < this->lf_index.size(); lpc++) {
+                        logline &line_to_update = this->lf_index[lpc];
+
+                        line_to_update.set_time_skew(true);
+                        line_to_update.set_time(second_to_last.get_time());
+                        line_to_update.set_millis(second_to_last.get_millis());
+                    }
+                }
             }
-        }
-    }
+            break;
+        case log_format::SCAN_NO_MATCH: {
+            logline::level_t last_level = logline::LEVEL_UNKNOWN;
+            time_t last_time = this->lf_index_time;
+            short last_millis = 0;
+            uint8_t last_mod = 0, last_opid = 0;
 
-    /* If the scanner didn't match, than we need to add it. */
-    if (!found) {
-        logline::level_t last_level  = logline::LEVEL_UNKNOWN;
-        time_t           last_time   = this->lf_index_time;
-        short            last_millis = 0;
+            if (!this->lf_index.empty()) {
+                logline &ll = this->lf_index.back();
 
-        if (!this->lf_index.empty()) {
-            logline &ll = this->lf_index.back();
-
-            /*
-             * Assume this line is part of the previous one(s) and copy the
-             * metadata over.
-             */
-            last_time   = ll.get_time();
-            last_millis = ll.get_millis();
-            if (this->lf_format.get() != NULL) {
-                last_level = (logline::level_t)
-                             (ll.get_level() | logline::LEVEL_CONTINUED);
+                /*
+                 * Assume this line is part of the previous one(s) and copy the
+                 * metadata over.
+                 */
+                last_time = ll.get_time();
+                last_millis = ll.get_millis();
+                if (this->lf_format.get() != NULL) {
+                    last_level = (logline::level_t)(ll.get_level_and_flags() |
+                        logline::LEVEL_CONTINUED);
+                }
+                last_mod = ll.get_module_id();
+                last_opid = ll.get_opid();
             }
+            this->lf_index.push_back(logline(offset,
+                                             last_time,
+                                             last_millis,
+                                             last_level,
+                                             last_mod,
+                                             last_opid));
+            break;
         }
-        this->lf_index.push_back(logline(offset,
-                                         last_time,
-                                         last_millis,
-                                         last_level));
+        case log_format::SCAN_INCOMPLETE:
+            break;
     }
 }
 
-bool logfile::rebuild_index(logfile_observer *lo)
+bool logfile::rebuild_index()
 throw (line_buffer::error, logfile::error)
 {
     bool        retval = false;
     struct stat st;
 
+    this->lf_activity.la_polls += 1;
+
     if (fstat(this->lf_line_buffer.get_fd(), &st) == -1) {
         throw error(this->lf_filename, errno);
     }
 
-    /* Check for new data based on the file size. */
-    if (this->lf_index_size < st.st_size) {
+    // Check the previous stat against the last to see if things are wonky.
+    if (this->lf_stat.st_size > st.st_size) {
+        log_info("truncated file detected, closing -- %s",
+                 this->lf_filename.c_str());
+        this->close();
+        return false;
+    }
+    else if (this->lf_line_buffer.is_data_available(this->lf_index_size, st.st_size)) {
+        this->lf_activity.la_reads += 1;
+
+        // We haven't reached the end of the file.  Note that we use the
+        // line buffer's notion of the file size since it may be compressed.
+        bool has_format = this->lf_format.get() != NULL;
+        struct rusage begin_rusage;
+        shared_buffer_ref sbr;
         off_t  last_off, off;
         line_value lv;
+        size_t begin_size = this->lf_index.size();
+        bool record_rusage = this->lf_index.size() == 1;
+
+        if (record_rusage) {
+            getrusage(RUSAGE_SELF, &begin_rusage);
+        }
 
         if (!this->lf_index.empty()) {
             off = this->lf_index.back().get_offset();
@@ -245,26 +310,67 @@ throw (line_buffer::error, logfile::error)
                 this->lf_index.pop_back();
             }
             this->lf_index.pop_back();
+
+            this->lf_line_buffer.clear();
+            if (!this->lf_index.empty()) {
+                off_t check_line_off = this->lf_index.back().get_offset();
+
+                if (!this->lf_line_buffer.read_line(check_line_off, sbr, &lv) ||
+                        off != check_line_off) {
+                    log_info("overwritten file detected, closing -- %s",
+                             this->lf_filename.c_str());
+                    this->close();
+                    return false;
+                }
+            }
         }
         else {
             off = 0;
         }
         last_off = off;
-        while (this->lf_line_buffer.read_line(off, lv)) {
-            char tmp = lv.lv_start[lv.lv_len];
+        if (this->lf_logline_observer != NULL) {
+            this->lf_logline_observer->logline_restart(*this);
+        }
+        while (this->lf_line_buffer.read_line(off, sbr, &lv)) {
+            size_t old_size = this->lf_index.size();
 
+            this->lf_longest_line = std::max(this->lf_longest_line, sbr.length());
             this->lf_partial_line = lv.lv_partial;
-            lv.lv_start[lv.lv_len] = '\0';
-            this->process_prefix(last_off, lv.lv_start, lv.lv_len);
-            lv.lv_start[lv.lv_len] = tmp;
+            this->process_prefix(last_off, sbr);
             last_off = off;
 
-            if (lo != NULL) {
-                lo->logfile_indexing(
+            for (logfile::iterator iter = this->begin() + old_size;
+                    iter != this->end(); ++iter) {
+                if (this->lf_logline_observer != NULL) {
+                    this->lf_logline_observer->logline_new_line(*this, iter, sbr);
+                }
+            }
+
+            if (this->lf_logfile_observer != NULL) {
+                this->lf_logfile_observer->logfile_indexing(
                     *this,
                     this->lf_line_buffer.get_read_offset(off),
                     st.st_size);
             }
+
+            if (!has_format && this->lf_format.get() != NULL) {
+                break;
+            }
+        }
+        if (this->lf_logline_observer != NULL) {
+            this->lf_logline_observer->logline_eof(*this);
+        }
+
+        if (record_rusage && (off - this->lf_index_size) > (500 * 1024)) {
+            struct rusage end_rusage;
+
+            getrusage(RUSAGE_SELF, &end_rusage);
+            rusagesub(end_rusage, begin_rusage, this->lf_activity.la_initial_index_rusage);
+            log_info("Resource usage for initial indexing of file: %s:%d-%d",
+                     this->lf_filename.c_str(),
+                     begin_size,
+                     this->lf_index.size());
+            log_rusage(LOG_LEVEL_INFO, this->lf_activity.la_initial_index_rusage);
         }
 
         /*
@@ -273,8 +379,7 @@ throw (line_buffer::error, logfile::error)
          * size.
          */
         this->lf_index_size = off;
-
-        // this->lf_line_buffer.invalidate();
+        this->lf_stat = st;
 
         retval = true;
     }
@@ -287,63 +392,6 @@ throw (line_buffer::error, logfile::error)
     return retval;
 }
 
-logfile_filter::type_t logfile::check_filter(iterator ll,
-                                             uint8_t generation,
-                                             const filter_stack_t &filters)
-{
-    logfile_filter::type_t retval;
-    uint8_t this_generation = ll->get_filter_generation();
-
-    if (this_generation == generation) {
-        return ll->get_filter_state();
-    }
-    else {
-        retval = logfile_filter::MAYBE;
-    }
-
-    string line_value;
-
-    for (size_t lpc = 0; lpc < filters.size(); lpc++) {
-        logfile_filter *filter = filters[lpc];
-        bool matched;
-
-        if (!filter->is_enabled())
-            continue;
-
-        if (line_value.empty())
-            this->read_line(ll, line_value);
-        matched = filter->matches(*ll, line_value);
-
-        switch (filter->get_type()) {
-        case logfile_filter::INCLUDE:
-            if (matched) {
-                // Always prefer including something.
-                retval = logfile_filter::INCLUDE;
-            }
-            else if (retval == logfile_filter::MAYBE) {
-                // Only exclude if we haven't made a decision yet.
-                retval = logfile_filter::EXCLUDE;
-            }
-            break;
-
-        case logfile_filter::EXCLUDE:
-            if (matched && retval == logfile_filter::MAYBE) {
-                // Only exclude if we haven't decide to include the line.
-                retval = logfile_filter::EXCLUDE;
-            }
-            break;
-
-        default:
-            require(0);
-            break;
-        }
-    }
-
-    ll->set_filter_state(generation, retval);
-
-    return retval;
-}
-
 void logfile::read_line(logfile::iterator ll, string &line_out)
 {
     try {
@@ -352,11 +400,8 @@ void logfile::read_line(logfile::iterator ll, string &line_out)
 
         line_out.clear();
         if (this->lf_line_buffer.read_line(off, sbr)) {
-            ostringstream stream;
-
             if (this->lf_format.get() != NULL) {
                 this->lf_format->get_subline(*ll, sbr);
-
             }
             line_out.append(sbr.get_data(), sbr.length());
         }
@@ -412,7 +457,9 @@ void logfile::read_full_message(logfile::iterator ll,
             /* ... */
         }
         ++ll;
-        max_lines -= 1;
+        if (max_lines != -1) {
+            max_lines -= 1;
+        }
     } while (ll != this->end() && ll->is_continued() &&
              (max_lines == -1 || max_lines > 0));
 
@@ -432,7 +479,7 @@ void logfile::read_full_message(logfile::iterator ll,
 
         if (this->lf_line_buffer.read_range(off, line_len, msg_out)) {
             if (this->lf_format.get() != NULL) {
-                this->lf_format->get_subline(*ll, msg_out);
+                this->lf_format->get_subline(*ll, msg_out, true);
             }
         }
         else {
@@ -443,3 +490,35 @@ void logfile::read_full_message(logfile::iterator ll,
         /* ... */
     }
 }
+
+void logfile::set_logline_observer(logline_observer *llo)
+{
+    this->lf_logline_observer = llo;
+    if (llo != NULL) {
+        this->reobserve_from(this->begin());
+    }
+}
+
+void logfile::reobserve_from(iterator iter)
+{
+    if (this->lf_logline_observer != NULL) {
+        for (; iter != this->end(); ++iter) {
+            off_t offset = std::distance(this->begin(), iter);
+            shared_buffer_ref sbr;
+
+            if (this->lf_logfile_observer != NULL) {
+                this->lf_logfile_observer->logfile_indexing(
+                        *this, offset, this->size());
+            }
+
+            this->read_line(iter, sbr);
+            this->lf_logline_observer->logline_new_line(*this, iter, sbr);
+        }
+        if (this->lf_logfile_observer != NULL) {
+            this->lf_logfile_observer->logfile_indexing(
+                    *this, this->size(), this->size());
+        }
+
+        this->lf_logline_observer->logline_eof(*this);
+    }
+}
