@@ -34,95 +34,24 @@
 #include "k_merge_tree.h"
 #include "lnav_util.hh"
 #include "log_accel.hh"
+#include "relative_time.hh"
 #include "logfile_sub_source.hh"
 
 using namespace std;
 
-#if 0
-/* XXX */
-class logfile_scrub_map {
-public:
-
-    static logfile_scrub_map &singleton()
-    {
-        static logfile_scrub_map s_lsm;
-
-        return s_lsm;
-    };
-
-    const pcre *include(logfile::format_t format) const
-    {
-        map<logfile::format_t, pcre *>::const_iterator iter;
-        pcre *retval = NULL;
-
-        if ((iter = this->lsm_include.find(format)) !=
-            this->lsm_include.end()) {
-            retval = iter->second;
-        }
-
-        return retval;
-    };
-
-    const pcre *exclude(logfile::format_t format) const
-    {
-        map<logfile::format_t, pcre *>::const_iterator iter;
-        pcre *retval = NULL;
-
-        if ((iter = this->lsm_exclude.find(format)) !=
-            this->lsm_exclude.end()) {
-            retval = iter->second;
-        }
-
-        return retval;
-    };
-
-private:
-
-    pcre *build_pcre(const char *pattern)
-    {
-        const char *errptr;
-        pcre *      retval;
-        int         eoff;
-
-        retval = pcre_compile(pattern, PCRE_CASELESS, &errptr, &eoff, NULL);
-        if (retval == NULL) {
-            throw errptr;
-        }
-
-        return retval;
-    };
-
-    logfile_scrub_map()
-    {
-        this->lsm_include[logfile::FORMAT_JBOSS] = this->
-                                                   build_pcre(
-            "\\d+-(\\d+-\\d+ \\d+:\\d+:\\d+),\\d+ [^ ]+( .*)");
-        this->lsm_exclude[logfile::FORMAT_JBOSS] = this->
-                                                   build_pcre("(?:"
-                                                              "\\[((?:[0-9a-zA-Z]+\\.)+)\\w+"
-                                                              "|\\[[\\w: ]+ ((?:[0-9a-zA-Z]+\\.)+)\\w+[^ \\]]+"
-                                                              "| ((?:[0-9a-zA-Z]+\\.)+)\\w+\\])");
-
-        this->lsm_include[logfile::FORMAT_SYSLOG] = this->
-                                                    build_pcre(
-            "(\\w+\\s[\\s\\d]\\d \\d+:\\d+:\\d+) \\w+( .*)");
-    };
-
-    map<logfile::format_t, pcre *> lsm_include;
-    map<logfile::format_t, pcre *> lsm_exclude;
-};
-#endif
-
-bookmark_type_t logfile_sub_source::BM_ERRORS;
-bookmark_type_t logfile_sub_source::BM_WARNINGS;
-bookmark_type_t logfile_sub_source::BM_FILES;
+bookmark_type_t logfile_sub_source::BM_ERRORS("error");
+bookmark_type_t logfile_sub_source::BM_WARNINGS("warning");
+bookmark_type_t logfile_sub_source::BM_FILES("");
 
 logfile_sub_source::logfile_sub_source()
     : lss_flags(0),
-      lss_filter_generation(1),
-      lss_filtered_count(0)
+      lss_token_file(NULL),
+      lss_min_log_level(logline::LEVEL_UNKNOWN),
+      lss_index_delegate(NULL),
+      lss_longest_line(0)
 {
     this->clear_line_size_cache();
+    this->clear_min_max_log_times();
 }
 
 logfile_sub_source::~logfile_sub_source()
@@ -131,15 +60,19 @@ logfile_sub_source::~logfile_sub_source()
 logfile *logfile_sub_source::find(const char *fn,
                                   content_line_t &line_base)
 {
-    std::vector<logfile_data>::iterator iter;
+    iterator iter;
     logfile *retval = NULL;
 
     line_base = content_line_t(0);
     for (iter = this->lss_files.begin();
          iter != this->lss_files.end() && retval == NULL;
          iter++) {
-        if (strcmp(iter->ld_file->get_filename().c_str(), fn) == 0) {
-            retval = iter->ld_file;
+        logfile_data &ld = *(*iter);
+        if (ld.get_file() == NULL) {
+            continue;
+        }
+        if (strcmp(ld.get_file()->get_filename().c_str(), fn) == 0) {
+            retval = ld.get_file();
         }
         else {
             line_base += content_line_t(MAX_LINES_PER_FILE);
@@ -151,15 +84,15 @@ logfile *logfile_sub_source::find(const char *fn,
 
 vis_line_t logfile_sub_source::find_from_time(const struct timeval &start)
 {
-    vector<indexed_content>::iterator lb;
+    vector<uint32_t>::iterator lb;
     vis_line_t retval(-1);
 
-    lb = lower_bound(this->lss_index.begin(),
-                     this->lss_index.end(),
+    lb = lower_bound(this->lss_filtered_index.begin(),
+                     this->lss_filtered_index.end(),
                      start,
-                     logline_cmp(*this));
-    if (lb != this->lss_index.end()) {
-        retval = vis_line_t(lb - this->lss_index.begin());
+                     filtered_logline_cmp(*this));
+    if (lb != this->lss_filtered_index.end()) {
+        retval = vis_line_t(lb - this->lss_filtered_index.begin());
     }
 
     return retval;
@@ -173,66 +106,93 @@ void logfile_sub_source::text_value_for_line(textview_curses &tc,
     content_line_t line(0);
 
     require(row >= 0);
-    require((size_t)row < this->lss_index.size());
+    require((size_t)row < this->lss_filtered_index.size());
 
     line = this->at(vis_line_t(row));
     this->lss_token_file   = this->find(line);
     this->lss_token_line   = this->lss_token_file->begin() + line;
-    this->lss_token_offset = 0;
-    this->lss_scrub_len    = 0;
 
     if (raw) {
         this->lss_token_file->read_line(this->lss_token_line, value_out);
         return;
     }
 
+    this->lss_token_attrs.clear();
+    this->lss_token_values.clear();
     this->lss_share_manager.invalidate_refs();
     this->lss_token_value =
         this->lss_token_file->read_line(this->lss_token_line);
 
-    this->lss_token_date_end = 0;
+    log_format *format = this->lss_token_file->get_format();
+
     value_out = this->lss_token_value;
     if (this->lss_flags & F_SCRUB) {
-        log_format *lf = this->lss_token_file->get_format();
-
-        lf->scrub(value_out);
+        format->scrub(value_out);
     }
 
-    if (this->lss_token_file->is_time_adjusted() &&
-        !(this->lss_token_line->get_level() & logline::LEVEL_CONTINUED)) {
-        log_format *format = this->lss_token_file->get_format();
+    if (!this->lss_token_line->is_continued()) {
+        shared_buffer_ref sbr;
 
-        if (format->lf_date_time.dts_fmt_lock != -1) {
-            shared_buffer_ref sbr;
-            std::vector<logline_value> line_values;
-            struct timeval adjusted_time;
-            string_attrs_t sa;
-            char buffer[128];
-            const char *fmt;
-
-            fmt = std_time_fmt[format->lf_date_time.dts_fmt_lock];
-            adjusted_time = this->lss_token_line->get_timeval();
-            strftime(buffer, sizeof(buffer),
-                     fmt,
-                     gmtime(&adjusted_time.tv_sec));
-
-            sbr.share(this->lss_share_manager,
-                (char *)this->lss_token_value.c_str(), this->lss_token_value.size());
-            format->annotate(sbr, sa, line_values);
+        sbr.share(this->lss_share_manager,
+                  (char *)this->lss_token_value.c_str(), this->lss_token_value.size());
+        format->annotate(sbr, this->lss_token_attrs, this->lss_token_values);
 
+        if ((this->lss_token_file->is_time_adjusted() ||
+             format->lf_timestamp_flags & ETF_MACHINE_ORIENTED) &&
+            format->lf_date_time.dts_fmt_lock != -1) {
             struct line_range time_range;
 
-            time_range = find_string_attr_range(sa, &logline::L_TIMESTAMP);
-            if (time_range.lr_start != -1) {
+            time_range = find_string_attr_range(
+                this->lss_token_attrs, &logline::L_TIMESTAMP);
+            if (time_range.is_valid()) {
+                struct timeval adjusted_time;
+                struct tm adjusted_tm;
+                char buffer[128];
+                const char *fmt;
+
+                if (format->lf_timestamp_flags & ETF_MACHINE_ORIENTED) {
+                    format->lf_date_time.convert_to_timeval(
+                        &this->lss_token_value.c_str()[time_range.lr_start],
+                        time_range.length(),
+                        adjusted_time);
+                    fmt = "%Y-%m-%d %H:%M:%S.";
+                }
+                else {
+                    fmt = PTIMEC_FORMAT_STR[format->lf_date_time.dts_fmt_lock];
+                    adjusted_time = this->lss_token_line->get_timeval();
+                }
+                strftime(buffer, sizeof(buffer),
+                         fmt,
+                         gmtime_r(&adjusted_time.tv_sec, &adjusted_tm));
+
+                if (format->lf_timestamp_flags & ETF_MACHINE_ORIENTED) {
+                    size_t buffer_len = strlen(buffer);
+
+                    snprintf(&buffer[buffer_len], sizeof(buffer) - buffer_len,
+                             "%06d",
+                             adjusted_time.tv_usec);
+                }
+
                 const char *last = value_out.c_str();
-                int len = strlen(buffer);
+                ssize_t len = strlen(buffer);
 
                 if ((last[time_range.lr_start + len] == '.' ||
-                    last[time_range.lr_start + len] == ',') &&
+                     last[time_range.lr_start + len] == ',') &&
                     len + 4 <= time_range.length()) {
-                    snprintf(&buffer[len], sizeof(buffer) - len,
-                             ".%03d",
-                             this->lss_token_line->get_millis());
+                    len = snprintf(&buffer[len], sizeof(buffer) - len,
+                                   ".%03d",
+                                   this->lss_token_line->get_millis());
+                }
+
+                if (len > time_range.length()) {
+                    ssize_t padding = len - time_range.length();
+
+                    value_out.insert(time_range.lr_start,
+                                     padding,
+                                     ' ');
+                    shift_string_attrs(this->lss_token_attrs,
+                                       time_range.lr_start + 1,
+                                       padding);
                 }
                 value_out.replace(time_range.lr_start,
                                   strlen(buffer),
@@ -241,6 +201,11 @@ void logfile_sub_source::text_value_for_line(textview_curses &tc,
         }
     }
 
+    if (this->lss_files.size() > 1) {
+        // Insert space for the file markers.
+        value_out.insert(0, 1, ' ');
+    }
+
     if (this->lss_flags & F_TIME_OFFSET) {
         int64_t start_millis, curr_millis;
 
@@ -255,47 +220,10 @@ void logfile_sub_source::text_value_for_line(textview_curses &tc,
         curr_millis = this->lss_token_line->get_time_in_millis();
         int64_t diff = curr_millis - start_millis;
 
-        /* 24h22m33s111 */
-
-        static struct rel_interval {
-            long long   length;
-            const char *format;
-            const char *symbol;
-        } intervals[] = {
-            { 1000, "%03qd%s", ""  },
-            {   60, "%qd%s",   "s" },
-            {   60, "%qd%s",   "m" },
-            {    0, "%qd%s",   "h" },
-            {    0, NULL, NULL }
-        };
-
-        struct rel_interval *curr_interval;
-        int rel_length = 0;
-
         value_out = "|" + value_out;
-        for (curr_interval = intervals; curr_interval->symbol != NULL;
-             curr_interval++) {
-            long long amount;
-            char      segment[32];
-
-            if (curr_interval->length) {
-                amount = diff % curr_interval->length;
-                diff   = diff / curr_interval->length;
-            }
-            else {
-                amount = diff;
-                diff   = 0;
-            }
-
-            if (!amount && !diff) {
-                break;
-            }
-
-            snprintf(segment, sizeof(segment), curr_interval->format, amount,
-                     curr_interval->symbol);
-            rel_length += strlen(segment);
-            value_out   = segment + value_out;
-        }
+        string relstr;
+        size_t rel_length = str2reltime(diff, relstr);
+        value_out.insert(0, relstr);
         if (rel_length < 12) {
             value_out.insert(0, 12 - rel_length, ' ');
         }
@@ -312,7 +240,8 @@ void logfile_sub_source::text_attrs_for_line(textview_curses &lv,
     int time_offset_end = 0;
     int attrs           = 0;
 
-    switch (this->lss_token_line->get_level() & ~logline::LEVEL__FLAGS) {
+    value_out = this->lss_token_attrs;
+    switch (this->lss_token_line->get_msg_level()) {
     case logline::LEVEL_FATAL:
     case logline::LEVEL_CRITICAL:
     case logline::LEVEL_ERROR:
@@ -328,7 +257,7 @@ void logfile_sub_source::text_attrs_for_line(textview_curses &lv,
         break;
     }
 
-    if ((row + 1) < (int)this->lss_index.size()) {
+    if ((row + 1) < (int)this->lss_filtered_index.size()) {
         next_line = this->find_line(this->at(vis_line_t(row + 1)));
     }
 
@@ -338,22 +267,62 @@ void logfile_sub_source::text_attrs_for_line(textview_curses &lv,
         attrs |= A_UNDERLINE;
     }
 
-    log_format *format = this->lss_token_file->get_format();
-    std::vector<logline_value> line_values;
-    
-    if (!(this->lss_token_line->get_level() & logline::LEVEL_CONTINUED)) {
-        shared_buffer_ref sbr;
+    std::vector<logline_value> &line_values = this->lss_token_values;
 
-        sbr.share(this->lss_share_manager,
-            (char *)this->lss_token_value.c_str(), this->lss_token_value.size());
-        format->annotate(sbr, value_out, line_values);
-    }
+    lr.lr_start = 0;
+    lr.lr_end = this->lss_token_value.length();
+    value_out.push_back(string_attr(lr, &textview_curses::SA_ORIGINAL_LINE));
 
-    lr.lr_start = time_offset_end + this->lss_token_date_end;
+    lr.lr_start = time_offset_end;
     lr.lr_end   = -1;
 
     value_out.push_back(string_attr(lr, &view_curses::VC_STYLE, attrs));
 
+    for (vector<logline_value>::iterator lv_iter = line_values.begin();
+         lv_iter != line_values.end();
+         ++lv_iter) {
+        if (!lv_iter->lv_identifier || !lv_iter->lv_origin.is_valid()) {
+            continue;
+        }
+
+        int id_attrs = vc.attrs_for_ident(lv_iter->text_value(),
+                                          lv_iter->text_length());
+
+        value_out.push_back(string_attr(
+                lv_iter->lv_origin, &view_curses::VC_STYLE, id_attrs));
+    }
+
+    if (this->lss_files.size() > 1) {
+        shift_string_attrs(value_out, 0, 1);
+
+        lr.lr_start = 0;
+        lr.lr_end = 1;
+        {
+            vis_bookmarks &bm = lv.get_bookmarks();
+            bookmark_vector<vis_line_t> &bv = bm[&BM_FILES];
+            bool is_first_for_file = binary_search(
+                    bv.begin(), bv.end(), vis_line_t(row));
+            bool is_last_for_file = binary_search(
+                    bv.begin(), bv.end(), vis_line_t(row + 1));
+            chtype graph = ACS_VLINE;
+            if (is_first_for_file) {
+                if (is_last_for_file) {
+                    graph = ACS_DIAMOND;
+                }
+                else {
+                    graph = ACS_ULCORNER;
+                }
+            }
+            else if (is_last_for_file) {
+                graph = ACS_LLCORNER;
+            }
+            value_out.push_back(
+                    string_attr(lr, &view_curses::VC_GRAPHIC, graph));
+        }
+        value_out.push_back(string_attr(lr, &view_curses::VC_STYLE, vc.attrs_for_ident(
+                this->lss_token_file->get_filename())));
+    }
+
     if (this->lss_flags & F_TIME_OFFSET) {
         time_offset_end = 13;
         lr.lr_start     = 0;
@@ -425,7 +394,7 @@ void logfile_sub_source::text_attrs_for_line(textview_curses &lv,
         }
     }
     else if ((((this->lss_token_line->get_time() / (5 * 60)) % 2) == 0) &&
-             !(this->lss_token_line->get_level() & logline::LEVEL_CONTINUED)) {
+             !this->lss_token_line->is_continued()) {
         struct line_range time_range = find_string_attr_range(
             value_out, &logline::L_TIMESTAMP);
 
@@ -434,11 +403,21 @@ void logfile_sub_source::text_attrs_for_line(textview_curses &lv,
             value_out.push_back(string_attr(time_range, &view_curses::VC_STYLE, attrs));
         }
     }
+
+    if (this->lss_token_line->is_time_skewed()) {
+        struct line_range time_range = find_string_attr_range(
+            value_out, &logline::L_TIMESTAMP);
+
+        if (time_range.lr_end != -1) {
+            attrs = vc.attrs_for_role(view_colors::VCR_SKEWED_TIME);
+            value_out.push_back(string_attr(time_range, &view_curses::VC_STYLE, attrs));
+        }
+    }
 }
 
-bool logfile_sub_source::rebuild_index(observer *obs, bool force)
+bool logfile_sub_source::rebuild_index(bool force)
 {
-    std::vector<logfile_data>::iterator iter;
+    iterator iter;
     size_t total_lines = 0;
     bool retval = force;
     int file_count = 0;
@@ -446,54 +425,54 @@ bool logfile_sub_source::rebuild_index(observer *obs, bool force)
     for (iter = this->lss_files.begin();
          iter != this->lss_files.end();
          iter++) {
-        if (iter->ld_file == NULL) {
-            if (iter->ld_lines_indexed > 0) {
+        if ((*iter)->get_file() == NULL) {
+            if ((*iter)->ld_lines_indexed > 0) {
                 force  = true;
                 retval = true;
             }
         }
         else {
-            if (iter->ld_file->rebuild_index(obs)) {
+            if ((*iter)->get_file()->rebuild_index()) {
                 retval = true;
             }
             file_count += 1;
-            total_lines += iter->ld_file->size();
+            total_lines += (*iter)->get_file()->size();
         }
     }
     if (force) {
         for (iter = this->lss_files.begin();
              iter != this->lss_files.end();
              iter++) {
-            iter->ld_lines_indexed = 0;
+            (*iter)->ld_lines_indexed = 0;
         }
+
+        this->lss_index.clear();
+        this->lss_filtered_index.clear();
+        this->lss_longest_line = 0;
     }
 
     if (retval || force) {
-        size_t index_size = 0;
-
-        if (force) {
-            this->lss_index.clear();
-            this->lss_filtered_count = 0;
-        }
+        size_t index_size = 0, start_size = this->lss_index.size();
 
         kmerge_tree_c<logline, logfile_data, logfile::iterator> merge(file_count);
 
         for (iter = this->lss_files.begin();
              iter != this->lss_files.end();
              iter++) {
-            if (iter->ld_file == NULL)
+            logfile_data *ld = *iter;
+            logfile *lf = ld->get_file();
+            if (lf == NULL) {
                 continue;
+            }
 
-            merge.add(&(*iter),
-                      iter->ld_file->begin() + iter->ld_lines_indexed,
-                      iter->ld_file->end());
-            index_size += iter->ld_file->size();
+            merge.add(ld,
+                      lf->begin() + ld->ld_lines_indexed,
+                      lf->end());
+            index_size += lf->size();
+            this->lss_longest_line = std::max(this->lss_longest_line, lf->get_longest_line_length());
         }
 
-        this->lss_index.reserve(index_size);
-
-        logfile_filter::type_t action_for_prev_line = logfile_filter::MAYBE;
-        logfile_data *last_owner = NULL;
+        this->lss_index.reset();
 
         merge.execute();
         for (;;) {
@@ -504,81 +483,63 @@ bool logfile_sub_source::rebuild_index(observer *obs, bool force)
                 break;
             }
 
-            int file_index = ld - &(*this->lss_files.begin());
-            int line_index = lf_iter - ld->ld_file->begin();
+            int file_index = ld->ld_file_index;
+            int line_index = lf_iter - ld->get_file()->begin();
 
             content_line_t con_line(file_index * MAX_LINES_PER_FILE +
                                     line_index);
 
-            if (!(lf_iter->get_level() & logline::LEVEL_CONTINUED)) {
-                if (last_owner != NULL) {
-                    if (action_for_prev_line != logfile_filter::EXCLUDE) {
-                        // Append all of the lines from the previous message.
-                        while (last_owner->ld_indexing.ld_start <=
-                           last_owner->ld_indexing.ld_last) {
-                            this->lss_index.push_back(last_owner->ld_indexing.ld_start);
-                            ++last_owner->ld_indexing.ld_start;
-                        }
-                    }
-                    else {
-                        this->lss_filtered_count += 1;
-                    }
-                }
-                ld->ld_indexing.ld_start = con_line;
-                action_for_prev_line = logfile_filter::MAYBE;
+            off_t insert_point = this->lss_index.merge_value(
+                    con_line, logline_cmp(*this));
+            if (insert_point < (off_t)start_size) {
+                start_size = 0;
+                this->lss_filtered_index.clear();
             }
 
-            if (obs != NULL) {
-                obs->logfile_sub_source_filtering(
-                    *this,
-                    vis_line_t(this->lss_index.size()),
-                    total_lines);
-            }
-
-            ld->ld_indexing.ld_last = con_line;
-            if (action_for_prev_line != logfile_filter::INCLUDE) {
-                // We haven't decided to include yet, so check the filter again.
-                // Once we decide to include a sub-line, the whole log message will
-                // always be included.
-                logfile_filter::type_t action_for_this_line = ld->ld_file->check_filter(
-                    lf_iter, this->lss_filter_generation, this->lss_filters);
-
-                // Only record the filter results for this line if it's not decisive.
-                if (action_for_this_line != logfile_filter::MAYBE) {
-                    action_for_prev_line = action_for_this_line;
-                }
-            }
-
-            last_owner = ld;
-
             merge.next();
         }
-        if (obs != NULL) {
-            obs->logfile_sub_source_filtering(*this,
-                                              vis_line_t(total_lines - 1),
-                                              total_lines);
-        }
-
-        if (last_owner != NULL) {
-            if (action_for_prev_line != logfile_filter::EXCLUDE) {
-                while (last_owner->ld_indexing.ld_start <=
-                   last_owner->ld_indexing.ld_last) {
-                    this->lss_index.push_back(last_owner->ld_indexing.ld_start);
-                    ++last_owner->ld_indexing.ld_start;
-                }
-            }
-            else {
-                this->lss_filtered_count += 1;
-            }
-        }
 
         for (iter = this->lss_files.begin();
              iter != this->lss_files.end();
              iter++) {
-            if (iter->ld_file == NULL)
+            if ((*iter)->get_file() == NULL)
                 continue;
 
-            iter->ld_lines_indexed = iter->ld_file->size();
+            (*iter)->ld_lines_indexed = (*iter)->get_file()->size();
+        }
+
+        this->lss_index.finish();
+
+        this->lss_filtered_index.reserve(this->lss_index.size());
+
+        uint32_t filter_in_mask, filter_out_mask;
+        this->get_filters().get_enabled_mask(filter_in_mask, filter_out_mask);
+
+        if (start_size == 0 && this->lss_index_delegate != NULL) {
+            this->lss_index_delegate->index_start(*this);
+        }
+
+        for (size_t index_index = start_size;
+             index_index < this->lss_index.size();
+             index_index++) {
+            content_line_t cl = (content_line_t) this->lss_index[index_index];
+            uint64_t line_number;
+            logfile_data *ld = this->find_data(cl, line_number);
+            logfile::iterator line_iter = ld->get_file()->begin() + line_number;
+
+            if (!ld->ld_filter_state.excluded(filter_in_mask, filter_out_mask,
+                    line_number) && this->check_extra_filters(*line_iter)) {
+                this->lss_filtered_index.push_back(index_index);
+                if (this->lss_index_delegate != NULL) {
+                    logfile *lf = ld->get_file();
+                    this->lss_index_delegate->index_line(
+                            *this, lf, lf->begin() + line_number);
+                }
+            }
+        }
+
+        if (this->lss_index_delegate != NULL) {
+            this->lss_index_delegate->index_complete(*this);
         }
     }
 
@@ -601,7 +562,7 @@ void logfile_sub_source::text_update_marks(vis_bookmarks &bm)
         bm[iter->first].clear();
     }
 
-    for (; vl < (int)this->lss_index.size(); ++vl) {
+    for (; vl < (int)this->lss_filtered_index.size(); ++vl) {
         const content_line_t orig_cl = this->at(vl);
         content_line_t cl = orig_cl;
         logfile *      lf;
@@ -629,19 +590,22 @@ void logfile_sub_source::text_update_marks(vis_bookmarks &bm)
             bm[&BM_FILES].insert_once(vl);
         }
 
-        switch ((*lf)[cl].get_level() & ~logline::LEVEL_MARK) {
-        case logline::LEVEL_WARNING:
-            bm[&BM_WARNINGS].insert_once(vl);
-            break;
+        logfile::iterator line_iter = lf->begin() + cl;
+        if (!line_iter->is_continued()) {
+            switch (line_iter->get_msg_level()) {
+                case logline::LEVEL_WARNING:
+                    bm[&BM_WARNINGS].insert_once(vl);
+                    break;
 
-        case logline::LEVEL_FATAL:
-        case logline::LEVEL_ERROR:
-        case logline::LEVEL_CRITICAL:
-            bm[&BM_ERRORS].insert_once(vl);
-            break;
+                case logline::LEVEL_FATAL:
+                case logline::LEVEL_ERROR:
+                case logline::LEVEL_CRITICAL:
+                    bm[&BM_ERRORS].insert_once(vl);
+                    break;
 
-        default:
-            break;
+                default:
+                    break;
+            }
         }
 
         last_file = lf;
@@ -671,70 +635,45 @@ log_accel::direction_t logfile_sub_source::get_line_accel_direction(
     return la.get_direction();
 }
 
-#if 0
-void logfile_sub_source::handle_scroll(listview_curses *lc)
+void logfile_sub_source::text_filters_changed()
 {
-    printf("hello, world!\n");
-    return;
+    for (iterator iter = this->begin(); iter != this->end(); ++iter) {
+        logfile_data *ld = *iter;
+        logfile *lf = ld->get_file();
 
-    vis_line_t top = lc->get_top();
-
-    if (this->lss_index.empty()) {
-        this->lss_top_time    = -1;
-        this->lss_bottom_time = -1;
+        if (lf != NULL) {
+            ld->ld_filter_state.clear_deleted_filter_state();
+            lf->reobserve_from(lf->begin() + ld->ld_filter_state.get_min_count(lf->size()));
+        }
     }
-    else {
-        time_t     top_day, bottom_day, today, yesterday, now = time(NULL);
-        vis_line_t bottom(0), height(0);
 
-        unsigned long width;
-        char          status[32];
-        logline *     ll;
+    uint32_t filtered_in_mask, filtered_out_mask;
 
-        today     = day_num(now);
-        yesterday = today - 1;
+    this->get_filters().get_enabled_mask(filtered_in_mask, filtered_out_mask);
 
-        lc->get_dimensions(height, width);
+    if (this->lss_index_delegate != NULL) {
+        this->lss_index_delegate->index_start(*this);
+    }
 
-        ll = this->find_line(this->lss_index[top]);
-        this->lss_top_time = ll->get_time();
+    this->lss_filtered_index.clear();
+    for (size_t index_index = 0; index_index < this->lss_index.size(); index_index++) {
+        content_line_t cl = (content_line_t) this->lss_index[index_index];
+        uint64_t line_number;
+        logfile_data *ld = this->find_data(cl, line_number);
+        logfile::iterator line_iter = ld->get_file()->begin() + line_number;
 
-        bottom = min(top + height - vis_line_t(1),
-                     vis_line_t(this->lss_index.size() - 1));
-
-        ll = this->find_line(this->lss_index[bottom]);
-        this->lss_bottom_time = ll->get_time();
-
-        top_day    = day_num(this->lss_top_time);
-        bottom_day = day_num(this->lss_bottom_time);
-        if (top_day == today) {
-            snprintf(status, sizeof(status), "Today");
-        }
-        else if (top_day == yesterday) {
-            snprintf(status, sizeof(status), "Yesterday");
-        }
-        else {
-            strftime(status, sizeof(status),
-                     "%a %b %d",
-                     gmtime(&this->lss_top_time));
-        }
-        if (top_day != bottom_day) {
-            int len = strlen(status);
-
-            if (bottom_day == today) {
-                snprintf(&status[len], sizeof(status) - len, " - Today");
-            }
-            else if (bottom_day == yesterday) {
-                snprintf(&status[len], sizeof(status) - len, " - Yesterday");
-            }
-            else {
-                strftime(&status[len], sizeof(status) - len,
-                         " - %b %d",
-                         gmtime(&this->lss_bottom_time));
+        if (!ld->ld_filter_state.excluded(filtered_in_mask, filtered_out_mask,
+                line_number) && this->check_extra_filters(*line_iter)) {
+            this->lss_filtered_index.push_back(index_index);
+            if (this->lss_index_delegate != NULL) {
+                logfile *lf = ld->get_file();
+                this->lss_index_delegate->index_line(
+                        *this, lf, lf->begin() + line_number);
             }
         }
+    }
 
-        this->lss_date_field.set_value(string(status));
+    if (this->lss_index_delegate != NULL) {
+        this->lss_index_delegate->index_complete(*this);
     }
 }
-#endif
