@@ -48,6 +48,7 @@
 #include "lnav_util.hh"
 #include "lnav_config.hh"
 #include "session_data.hh"
+#include "command_executor.hh"
 
 using namespace std;
 
@@ -113,33 +114,37 @@ static bool bind_line(sqlite3 *db,
     if (sqlite3_bind_text(stmt, 1,
                           timestamp, strlen(timestamp),
                           SQLITE_TRANSIENT) != SQLITE_OK) {
-        fprintf(stderr, "error: could not bind log time -- %s\n",
+        log_error("could not bind log time -- %s\n",
                 sqlite3_errmsg(db));
         return false;
     }
 
-    const std::string &format_name = lf->get_format()->get_name();
+    intern_string_t format_name = lf->get_format()->get_name();
 
     if (sqlite3_bind_text(stmt, 2,
-                          format_name.c_str(), format_name.length(),
+                          format_name.get(), format_name.size(),
                           SQLITE_TRANSIENT) != SQLITE_OK) {
-        fprintf(stderr, "error: could not bind log format -- %s\n",
+        log_error("could not bind log format -- %s\n",
                     sqlite3_errmsg(db));
         return false;
     }
 
-    std::string line_hash = hash_string(lf->read_line(line_iter));
+    shared_buffer_ref sbr;
+    lf->read_line(line_iter, sbr);
+    std::string line_hash = hash_bytes(sbr.get_data(), sbr.length(),
+                                       &cl, sizeof(cl),
+                                       NULL);
 
     if (sqlite3_bind_text(stmt, 3,
                           line_hash.c_str(), line_hash.length(),
                           SQLITE_TRANSIENT) != SQLITE_OK) {
-        fprintf(stderr, "error: could not bind log hash -- %s\n",
+        log_error("could not bind log hash -- %s\n",
                 sqlite3_errmsg(db));
         return false;
     }
 
     if (sqlite3_bind_int64(stmt, 4, session_time) != SQLITE_OK) {
-        fprintf(stderr, "error: could not bind session time -- %s\n",
+        log_error("could not bind session time -- %s\n",
                 sqlite3_errmsg(db));
         return false;
     }
@@ -222,8 +227,8 @@ static void cleanup_session_data(void)
         }
         else {
             if (remove(front.sfi_path.c_str()) != 0) {
-                fprintf(stderr,
-                        "error: Unable to remove session file: %s -- %s\n",
+                log_error(
+                        "Unable to remove session file: %s -- %s\n",
                         front.sfi_path.c_str(),
                         strerror(errno));
             }
@@ -238,8 +243,8 @@ static void cleanup_session_data(void)
         const session_file_info &front = session_info_list.front();
 
         if (remove(front.sfi_path.c_str()) != 0) {
-            fprintf(stderr,
-                    "error: Unable to remove session file: %s -- %s\n",
+            log_error(
+                    "Unable to remove session file: %s -- %s\n",
                     front.sfi_path.c_str(),
                     strerror(errno));
         }
@@ -257,9 +262,11 @@ void init_session(void)
 
     context.Init(0, 0);
     hash_updater updater(&context);
-    for_each(lnav_data.ld_file_names.begin(),
-             lnav_data.ld_file_names.end(),
-             object_field(updater, &pair<string, int>::first));
+    for (map<string, logfile_open_options>::iterator iter = lnav_data.ld_file_names.begin();
+         iter != lnav_data.ld_file_names.end();
+         ++iter) {
+        updater(iter->first);
+    }
     context.Final(hash.out(0), hash.out(1));
 
     lnav_data.ld_session_id = hash.to_string();
@@ -323,8 +330,8 @@ void scan_sessions(void)
         const std::string &name = session_file_names.front().second;
 
         if (remove(name.c_str()) != 0) {
-            fprintf(stderr,
-                    "error: Unable to remove session: %s -- %s\n",
+            log_error(
+                    "Unable to remove session: %s -- %s\n",
                     name.c_str(),
                     strerror(errno));
         }
@@ -365,8 +372,8 @@ static void load_time_bookmarks(void)
                            -1,
                            stmt.out(),
                            NULL) != SQLITE_OK) {
-        fprintf(stderr,
-                "error: could not prepare bookmark select statemnt -- %s\n",
+        log_error(
+                "could not prepare bookmark select statemnt -- %s\n",
                 sqlite3_errmsg(db));
         return;
     }
@@ -375,7 +382,7 @@ static void load_time_bookmarks(void)
          file_iter != lnav_data.ld_log_source.end();
          ++file_iter) {
         char low_timestamp[64], high_timestamp[64];
-        logfile *lf = file_iter->ld_file;
+        logfile *lf = (*file_iter)->get_file();
         content_line_t base_content_line;
 
         if (lf == NULL)
@@ -389,7 +396,7 @@ static void load_time_bookmarks(void)
                      lf->original_line_time(line_iter), 'T');
 
         if (sqlite3_bind_int64(stmt.in(), 1, lnav_data.ld_session_load_time) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind session time -- %s\n",
+            log_error("could not bind session time -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
@@ -397,7 +404,7 @@ static void load_time_bookmarks(void)
         if (sqlite3_bind_text(stmt.in(), 2,
                               low_timestamp, strlen(low_timestamp),
                               SQLITE_TRANSIENT) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind low log time -- %s\n",
+            log_error("could not bind low log time -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
@@ -410,17 +417,17 @@ static void load_time_bookmarks(void)
         if (sqlite3_bind_text(stmt.in(), 3,
                               high_timestamp, strlen(high_timestamp),
                               SQLITE_TRANSIENT) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind high log time -- %s\n",
+            log_error("could not bind high log time -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
 
-        const std::string &format_name = lf->get_format()->get_name();
+        intern_string_t format_name = lf->get_format()->get_name();
 
         if (sqlite3_bind_text(stmt.in(), 4,
-                              format_name.c_str(), format_name.length(),
+                              format_name.get(), format_name.size(),
                               SQLITE_TRANSIENT) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind log format -- %s\n",
+            log_error("could not bind log format -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
@@ -445,7 +452,7 @@ static void load_time_bookmarks(void)
                 const char *part_name = (const char *)sqlite3_column_text(stmt.in(), 4);
                 int64_t mark_time = sqlite3_column_int64(stmt.in(), 3);
                 struct timeval log_tv;
-                struct tm log_tm;
+                struct exttm log_tm;
 
                 if (last_mark_time == -1) {
                     last_mark_time = mark_time;
@@ -459,7 +466,7 @@ static void load_time_bookmarks(void)
                     continue;
                 }
 
-                if (!dts.scan(log_time, NULL, &log_tm, log_tv)) {
+                if (!dts.scan(log_time, strlen(log_time), NULL, &log_tm, log_tv)) {
                     continue;
                 }
 
@@ -474,9 +481,14 @@ static void load_time_bookmarks(void)
                         break;
                     }
 
-                    lf->read_line(line_iter, line);
+                    shared_buffer_ref sbr;
+                    content_line_t cl = content_line_t(std::distance(lf->begin(), line_iter));
+                    lf->read_line(line_iter, sbr);
+
+                    string line_hash = hash_bytes(sbr.get_data(), sbr.length(),
+                                                  &cl, sizeof(cl),
+                                                  NULL);
 
-                    string line_hash = hash_string(line);
                     if (line_hash == log_hash) {
                         content_line_t line_cl = content_line_t(
                             base_content_line + std::distance(lf->begin(), line_iter));
@@ -504,8 +516,8 @@ static void load_time_bookmarks(void)
                     const char *errmsg;
 
                     errmsg = sqlite3_errmsg(lnav_data.ld_db);
-                    fprintf(stderr,
-                            "error: bookmark select error: code %d -- %s\n",
+                    log_error(
+                            "bookmark select error: code %d -- %s\n",
                             rc,
                             errmsg);
                     done = true;
@@ -524,8 +536,8 @@ static void load_time_bookmarks(void)
                            -1,
                            stmt.out(),
                            NULL) != SQLITE_OK) {
-        fprintf(stderr,
-                "error: could not prepare time_offset select statemnt -- %s\n",
+        log_error(
+                "could not prepare time_offset select statemnt -- %s\n",
                 sqlite3_errmsg(db));
         return;
     }
@@ -534,7 +546,7 @@ static void load_time_bookmarks(void)
          file_iter != lnav_data.ld_log_source.end();
          ++file_iter) {
         char low_timestamp[64], high_timestamp[64];
-        logfile *lf = file_iter->ld_file;
+        logfile *lf = (*file_iter)->get_file();
         content_line_t base_content_line;
 
         if (lf == NULL)
@@ -548,7 +560,7 @@ static void load_time_bookmarks(void)
                      lf->original_line_time(line_iter), 'T');
 
         if (sqlite3_bind_int64(stmt.in(), 1, lnav_data.ld_session_load_time) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind session time -- %s\n",
+            log_error("could not bind session time -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
@@ -556,7 +568,7 @@ static void load_time_bookmarks(void)
         if (sqlite3_bind_text(stmt.in(), 2,
                               low_timestamp, strlen(low_timestamp),
                               SQLITE_TRANSIENT) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind low log time -- %s\n",
+            log_error("could not bind low log time -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
@@ -569,17 +581,17 @@ static void load_time_bookmarks(void)
         if (sqlite3_bind_text(stmt.in(), 3,
                               high_timestamp, strlen(high_timestamp),
                               SQLITE_TRANSIENT) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind high log time -- %s\n",
+            log_error("could not bind high log time -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
 
-        const std::string &format_name = lf->get_format()->get_name();
+        intern_string_t format_name = lf->get_format()->get_name();
 
         if (sqlite3_bind_text(stmt.in(), 4,
-                              format_name.c_str(), format_name.length(),
+                              format_name.get(), format_name.size(),
                               SQLITE_TRANSIENT) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind log format -- %s\n",
+            log_error("could not bind log format -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
@@ -603,7 +615,7 @@ static void load_time_bookmarks(void)
                 const char *log_hash = (const char *)sqlite3_column_text(stmt.in(), 2);
                 int64_t mark_time = sqlite3_column_int64(stmt.in(), 3);
                 struct timeval log_tv;
-                struct tm log_tm;
+                struct exttm log_tm;
 
                 if (last_mark_time == -1) {
                     last_mark_time = mark_time;
@@ -617,7 +629,7 @@ static void load_time_bookmarks(void)
                     continue;
                 }
 
-                if (!dts.scan(log_time, NULL, &log_tm, log_tv)) {
+                if (!dts.scan(log_time, strlen(log_time), NULL, &log_tm, log_tv)) {
                     continue;
                 }
 
@@ -657,8 +669,8 @@ static void load_time_bookmarks(void)
                     const char *errmsg;
 
                     errmsg = sqlite3_errmsg(lnav_data.ld_db);
-                    fprintf(stderr,
-                            "error: bookmark select error: code %d -- %s\n",
+                    log_error(
+                            "bookmark select error: code %d -- %s\n",
                             rc,
                             errmsg);
                     done = true;
@@ -680,8 +692,6 @@ static int read_save_time(yajlpp_parse_context *ypc, long long value)
 
 static int read_time_offset(yajlpp_parse_context *ypc, int value)
 {
-    lnav_data.ld_log_source.set_time_offset(value);
-
     return 1;
 }
 
@@ -753,12 +763,28 @@ static int read_word_wrap(yajlpp_parse_context *ypc, int value)
 static int read_commands(yajlpp_parse_context *ypc, const unsigned char *str, size_t len)
 {
     std::string cmdline = std::string((const char *)str, len);
+    const char **         view_name;
+    int view_index;
 
+    view_name = find(lnav_view_strings,
+            lnav_view_strings + LNV__MAX,
+            ypc->get_path_fragment(-3));
+    view_index = view_name - lnav_view_strings;
+    bool active = ensure_view(&lnav_data.ld_views[view_index]);
     execute_command(cmdline);
+    if (!active) {
+        lnav_data.ld_view_stack.pop();
+    }
 
     return 1;
 }
 
+static struct json_path_handler view_handlers[] = {
+        json_path_handler("top_line"),
+
+        json_path_handler()
+};
+
 static struct json_path_handler view_info_handlers[] = {
     json_path_handler("/save-time",               read_save_time),
     json_path_handler("/time-offset",             read_time_offset),
@@ -766,7 +792,7 @@ static struct json_path_handler view_info_handlers[] = {
     json_path_handler("/views/([^/]+)/top_line",  read_top_line),
     json_path_handler("/views/([^/]+)/search",    read_last_search),
     json_path_handler("/views/([^/]+)/word_wrap", read_word_wrap),
-    json_path_handler("/commands#",               read_commands),
+    json_path_handler("/views/([^/]+)/commands#", read_commands),
 
     json_path_handler()
 };
@@ -798,7 +824,7 @@ void load_session(void)
     }
     else {
         unsigned char buffer[1024];
-        size_t        rc;
+        ssize_t        rc;
 
         log_info("loading session file: %s", view_info_name.c_str());
         while ((rc = read(fd, buffer, sizeof(buffer))) > 0) {
@@ -839,7 +865,7 @@ static void save_user_bookmarks(
 
         if (meta_iter == bm_meta.end()) {
             if (sqlite3_bind_text(stmt, 5, "", 0, SQLITE_TRANSIENT) != SQLITE_OK) {
-                fprintf(stderr, "error: could not bind log hash -- %s\n",
+                log_error("could not bind log hash -- %s\n",
                         sqlite3_errmsg(db));
                 return;
             }
@@ -849,15 +875,15 @@ static void save_user_bookmarks(
                                   meta_iter->second.bm_name.c_str(),
                                   meta_iter->second.bm_name.length(),
                                   SQLITE_TRANSIENT) != SQLITE_OK) {
-                fprintf(stderr, "error: could not bind log hash -- %s\n",
+                log_error("could not bind log hash -- %s\n",
                         sqlite3_errmsg(db));
                 return;
             }
         }
 
         if (sqlite3_step(stmt) != SQLITE_DONE) {
-            fprintf(stderr,
-                    "error: could not execute bookmark insert statement -- %s\n",
+            log_error(
+                    "could not execute bookmark insert statement -- %s\n",
                     sqlite3_errmsg(db));
             return;
         }
@@ -875,17 +901,17 @@ static void save_time_bookmarks(void)
     auto_mem<sqlite3_stmt> stmt(sqlite3_finalize);
 
     if (sqlite3_open(db_path.c_str(), db.out()) != SQLITE_OK) {
-        fprintf(stderr, "error: unable to open bookmark DB -- %s\n", db_path.c_str());
+        log_error("unable to open bookmark DB -- %s\n", db_path.c_str());
         return;
     }
 
     if (sqlite3_exec(db.in(), BOOKMARK_TABLE_DEF, NULL, NULL, errmsg.out()) != SQLITE_OK) {
-        fprintf(stderr, "error: unable to make bookmark table -- %s\n", errmsg.in());
+        log_error("unable to make bookmark table -- %s\n", errmsg.in());
         return;
     }
 
     if (sqlite3_exec(db.in(), "BEGIN TRANSACTION", NULL, NULL, errmsg.out()) != SQLITE_OK) {
-        fprintf(stderr, "error: unable to begin transaction -- %s\n", errmsg.in());
+        log_error("unable to begin transaction -- %s\n", errmsg.in());
         return;
     }
 
@@ -899,8 +925,8 @@ static void save_time_bookmarks(void)
                            -1,
                            stmt.out(),
                            NULL) != SQLITE_OK) {
-        fprintf(stderr,
-                "error: could not prepare bookmark delete statemnt -- %s\n",
+        log_error(
+                "could not prepare bookmark delete statemnt -- %s\n",
                 sqlite3_errmsg(db));
         return;
     }
@@ -914,8 +940,8 @@ static void save_time_bookmarks(void)
         }
 
         if (sqlite3_step(stmt.in()) != SQLITE_DONE) {
-            fprintf(stderr,
-                    "error: could not execute bookmark insert statement -- %s\n",
+            log_error(
+                    "could not execute bookmark insert statement -- %s\n",
                     sqlite3_errmsg(db));
             return;
         }
@@ -932,8 +958,8 @@ static void save_time_bookmarks(void)
                            -1,
                            stmt.out(),
                            NULL) != SQLITE_OK) {
-        fprintf(stderr,
-                "error: could not prepare bookmark replace statemnt -- %s\n",
+        log_error(
+                "could not prepare bookmark replace statemnt -- %s\n",
                 sqlite3_errmsg(db));
         return;
     }
@@ -944,7 +970,7 @@ static void save_time_bookmarks(void)
         for (file_iter = lnav_data.ld_log_source.begin();
              file_iter != lnav_data.ld_log_source.end();
              ++file_iter) {
-            logfile *lf = file_iter->ld_file;
+            logfile *lf = (*file_iter)->get_file();
             content_line_t base_content_line;
 
             if (lf == NULL)
@@ -960,14 +986,14 @@ static void save_time_bookmarks(void)
             }
 
             if (sqlite3_bind_null(stmt.in(), 5) != SQLITE_OK) {
-                fprintf(stderr, "error: could not bind log hash -- %s\n",
+                log_error("could not bind log hash -- %s\n",
                         sqlite3_errmsg(db.in()));
                 return;
             }
 
             if (sqlite3_step(stmt.in()) != SQLITE_DONE) {
-                fprintf(stderr,
-                        "error: could not execute bookmark insert statement -- %s\n",
+                log_error(
+                        "could not execute bookmark insert statement -- %s\n",
                         sqlite3_errmsg(db));
                 return;
             }
@@ -986,8 +1012,8 @@ static void save_time_bookmarks(void)
                            -1,
                            stmt.out(),
                            NULL) != SQLITE_OK) {
-        fprintf(stderr,
-                "error: could not prepare time_offset delete statemnt -- %s\n",
+        log_error(
+                "could not prepare time_offset delete statemnt -- %s\n",
                 sqlite3_errmsg(db));
         return;
     }
@@ -1001,8 +1027,8 @@ static void save_time_bookmarks(void)
         }
 
         if (sqlite3_step(stmt.in()) != SQLITE_DONE) {
-            fprintf(stderr,
-                    "error: could not execute bookmark insert statement -- %s\n",
+            log_error(
+                    "could not execute bookmark insert statement -- %s\n",
                     sqlite3_errmsg(db));
             return;
         }
@@ -1019,8 +1045,8 @@ static void save_time_bookmarks(void)
                            -1,
                            stmt.out(),
                            NULL) != SQLITE_OK) {
-        fprintf(stderr,
-                "error: could not prepare time_offset replace statemnt -- %s\n",
+        log_error(
+                "could not prepare time_offset replace statemnt -- %s\n",
                 sqlite3_errmsg(db));
         return;
     }
@@ -1031,7 +1057,7 @@ static void save_time_bookmarks(void)
         for (file_iter = lnav_data.ld_log_source.begin();
              file_iter != lnav_data.ld_log_source.end();
              ++file_iter) {
-            logfile *lf = file_iter->ld_file;
+            logfile *lf = (*file_iter)->get_file();
             content_line_t base_content_line;
 
             if (lf == NULL)
@@ -1045,20 +1071,20 @@ static void save_time_bookmarks(void)
             }
 
             if (sqlite3_bind_null(stmt.in(), 5) != SQLITE_OK) {
-                fprintf(stderr, "error: could not bind log hash -- %s\n",
+                log_error("could not bind log hash -- %s\n",
                         sqlite3_errmsg(db.in()));
                 return;
             }
 
             if (sqlite3_bind_null(stmt.in(), 6) != SQLITE_OK) {
-                fprintf(stderr, "error: could not bind log hash -- %s\n",
+                log_error("could not bind log hash -- %s\n",
                         sqlite3_errmsg(db.in()));
                 return;
             }
 
             if (sqlite3_step(stmt.in()) != SQLITE_DONE) {
-                fprintf(stderr,
-                        "error: could not execute bookmark insert statement -- %s\n",
+                log_error(
+                        "could not execute bookmark insert statement -- %s\n",
                         sqlite3_errmsg(db));
                 return;
             }
@@ -1072,10 +1098,10 @@ static void save_time_bookmarks(void)
          ++file_iter) {
         logfile::iterator line_iter;
 
-        if (file_iter->ld_file == NULL)
+        if ((*file_iter)->get_file() == NULL)
             continue;
 
-        logfile *lf = file_iter->ld_file;
+        logfile *lf = (*file_iter)->get_file();
 
         if (!lf->is_time_adjusted())
             continue;
@@ -1089,17 +1115,17 @@ static void save_time_bookmarks(void)
         if (sqlite3_bind_text(stmt.in(), 1,
                               timestamp, strlen(timestamp),
                               SQLITE_TRANSIENT) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind log time -- %s\n",
+            log_error("could not bind log time -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
 
-        const std::string &format_name = lf->get_format()->get_name();
+        intern_string_t format_name = lf->get_format()->get_name();
 
         if (sqlite3_bind_text(stmt.in(), 2,
-                              format_name.c_str(), format_name.length(),
+                              format_name.get(), format_name.size(),
                               SQLITE_TRANSIENT) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind log format -- %s\n",
+            log_error("could not bind log format -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
@@ -1109,13 +1135,13 @@ static void save_time_bookmarks(void)
         if (sqlite3_bind_text(stmt.in(), 3,
                               line_hash.c_str(), line_hash.length(),
                               SQLITE_TRANSIENT) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind log hash -- %s\n",
+            log_error("could not bind log hash -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
 
         if (sqlite3_bind_int64(stmt.in(), 4, lnav_data.ld_session_time) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind session time -- %s\n",
+            log_error("could not bind session time -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
@@ -1123,20 +1149,20 @@ static void save_time_bookmarks(void)
         struct timeval offset = lf->get_time_offset();
 
         if (sqlite3_bind_int64(stmt.in(), 5, offset.tv_sec) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind offset_sec -- %s\n",
+            log_error("could not bind offset_sec -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
 
         if (sqlite3_bind_int64(stmt.in(), 6, offset.tv_usec) != SQLITE_OK) {
-            fprintf(stderr, "error: could not bind offset_usec -- %s\n",
+            log_error("could not bind offset_usec -- %s\n",
                     sqlite3_errmsg(db.in()));
             return;
         }
 
         if (sqlite3_step(stmt.in()) != SQLITE_DONE) {
-            fprintf(stderr,
-                    "error: could not execute bookmark insert statement -- %s\n",
+            log_error(
+                    "could not execute bookmark insert statement -- %s\n",
                     sqlite3_errmsg(db));
             return;
         }
@@ -1145,12 +1171,12 @@ static void save_time_bookmarks(void)
     }
 
     if (sqlite3_exec(db.in(), "COMMIT", NULL, NULL, errmsg.out()) != SQLITE_OK) {
-        fprintf(stderr, "error: unable to begin transaction -- %s\n", errmsg.in());
+        log_error("unable to begin transaction -- %s\n", errmsg.in());
         return;
     }
 
     if (sqlite3_exec(db.in(), BOOKMARK_LRU_STMT, NULL, NULL, errmsg.out()) != SQLITE_OK) {
-        fprintf(stderr, "error: unable to delete old bookmarks -- %s\n", errmsg.in());
+        log_error("unable to delete old bookmarks -- %s\n", errmsg.in());
         return;
     }
 }
@@ -1183,7 +1209,6 @@ void save_session(void)
         perror("Unable to create yajl_gen object");
     }
     else {
-        yajl_gen_config(handle, yajl_gen_beautify, 1);
         yajl_gen_config(handle,
                         yajl_gen_print_callback, yajl_writer, file.in());
 
@@ -1201,10 +1226,11 @@ void save_session(void)
             {
                 yajlpp_array file_list(handle);
 
-                for_each(lnav_data.ld_file_names.begin(),
-                         lnav_data.ld_file_names.end(),
-                         object_field(file_list.gen,
-                                      &pair<string, int>::first));
+                for (map<string, logfile_open_options>::iterator iter = lnav_data.ld_file_names.begin();
+                     iter != lnav_data.ld_file_names.end();
+                     ++iter) {
+                    file_list.gen(iter->first);
+                }
             }
 
             root_map.gen("views");
@@ -1236,39 +1262,42 @@ void save_session(void)
 
                     view_map.gen("word_wrap");
                     view_map.gen(tc.get_word_wrap());
-                }
-            }
 
-            root_map.gen("commands");
-
-            {
-                filter_stack_t::const_iterator filter_iter;
-                logfile_sub_source &lss = lnav_data.ld_log_source;
-                const filter_stack_t &fs = lss.get_filters();
-
-                yajlpp_array cmd_array(handle);
-
-                for (filter_iter = fs.begin();
-                     filter_iter != fs.end();
-                     ++filter_iter) {
-                    if (!(*filter_iter)->is_enabled()) {
+                    text_sub_source *tss = tc.get_sub_source();
+                    if (tss == NULL) {
                         continue;
                     }
 
-                    cmd_array.gen((*filter_iter)->to_command());
-                }
+                    filter_stack::iterator filter_iter;
+                    filter_stack &fs = tss->get_filters();
 
-                textview_curses::highlight_map_t &hmap =
-                    lnav_data.ld_views[LNV_LOG].get_highlights();
-                textview_curses::highlight_map_t::iterator hl_iter;
+                    view_map.gen("commands");
+                    yajlpp_array cmd_array(handle);
 
-                for (hl_iter = hmap.begin();
-                     hl_iter != hmap.end();
-                     ++hl_iter) {
-                    if (hl_iter->first[0] == '$') {
-                        continue;
+                    for (filter_iter = fs.begin();
+                         filter_iter != fs.end();
+                         ++filter_iter) {
+                        if (!(*filter_iter)->is_enabled()) {
+                            continue;
+                        }
+
+                        cmd_array.gen((*filter_iter)->to_command());
+                    }
+
+                    if (lpc == LNV_LOG) {
+                        textview_curses::highlight_map_t &hmap =
+                                lnav_data.ld_views[LNV_LOG].get_highlights();
+                        textview_curses::highlight_map_t::iterator hl_iter;
+
+                        for (hl_iter = hmap.begin();
+                             hl_iter != hmap.end();
+                             ++hl_iter) {
+                            if (hl_iter->first[0] == '$') {
+                                continue;
+                            }
+                            cmd_array.gen("highlight " + hl_iter->first);
+                        }
                     }
-                    cmd_array.gen("highlight " + hl_iter->first);
                 }
             }
         }
@@ -1278,7 +1307,9 @@ void save_session(void)
 
         fclose(file.release());
 
-        rename(view_file_tmp_name.c_str(), view_file_name.c_str());
+        log_perror(rename(view_file_tmp_name.c_str(), view_file_name.c_str()));
+
+        log_debug("Saved session: %s", view_file_name.c_str());
     }
 }
 
@@ -1287,7 +1318,6 @@ void reset_session(void)
     textview_curses::highlight_map_t &hmap =
         lnav_data.ld_views[LNV_LOG].get_highlights();
     textview_curses::highlight_map_t::iterator hl_iter = hmap.begin();
-    logfile_sub_source &lss = lnav_data.ld_log_source;
 
     log_info("reset session: time=%d", lnav_data.ld_session_time);
 
@@ -1305,18 +1335,33 @@ void reset_session(void)
         }
     }
 
-    lnav_data.ld_log_source.clear_filters();
-
     logfile_sub_source::iterator file_iter;
 
     for (file_iter = lnav_data.ld_log_source.begin();
          file_iter != lnav_data.ld_log_source.end();
          ++file_iter) {
-        logfile *lf             = file_iter->ld_file;
+        logfile *lf             = (*file_iter)->get_file();
 
         lf->clear_time_offset();
     }
 
-    lss.get_user_bookmarks()[&textview_curses::BM_USER].clear();
-    lss.get_user_bookmarks()[&textview_curses::BM_PARTITION].clear();
+    lnav_data.ld_log_source.clear_min_max_log_times();
+
+    lnav_data.ld_log_source.get_user_bookmark_metadata().clear();
+
+    for (int lpc = 0; lpc < LNV__MAX; lpc++) {
+        textview_curses &tc = lnav_data.ld_views[lpc];
+        text_sub_source *tss = tc.get_sub_source();
+
+        if (tss == NULL) {
+            continue;
+        }
+        tss->get_filters().clear_filters();
+        tss->text_filters_changed();
+        tss->text_clear_marks(&textview_curses::BM_USER);
+        tc.get_bookmarks()[&textview_curses::BM_USER].clear();
+        tss->text_clear_marks(&textview_curses::BM_PARTITION);
+        tc.get_bookmarks()[&textview_curses::BM_PARTITION].clear();
+        tc.reload_data();
+    }
 }
