@@ -34,8 +34,11 @@
 #include <string>
 #include <algorithm>
 
+#include "auto_mem.hh"
 #include "lnav_log.hh"
 #include "view_curses.hh"
+#include "ansi_scrubber.hh"
+#include "lnav_config.hh"
 
 using namespace std;
 
@@ -48,8 +51,14 @@ const struct itimerval ui_periodic_timer::INTERVAL = {
 };
 
 ui_periodic_timer::ui_periodic_timer()
+        : upt_counter(0)
 {
-    signal(SIGALRM, ui_periodic_timer::sigalrm);
+    struct sigaction sa;
+
+    sa.sa_handler = ui_periodic_timer::sigalrm;
+    sa.sa_flags = SA_RESTART;
+    sigemptyset(&sa.sa_mask);
+    sigaction(SIGALRM, &sa, NULL);
     if (setitimer(ITIMER_REAL, &INTERVAL, NULL) == -1) {
         perror("setitimer");
     }
@@ -73,6 +82,23 @@ alerter &alerter::singleton() {
     return retval;
 }
 
+attr_line_t &attr_line_t::with_ansi_string(const char *str, ...)
+{
+    auto_mem<char> formatted_str;
+    va_list args;
+
+    va_start(args, str);
+    vasprintf(formatted_str.out(), str, args);
+    va_end(args);
+
+    if (formatted_str != NULL) {
+        this->al_string = formatted_str;
+        scrub_ansi_string(this->al_string, this->al_attrs);
+    }
+
+    return *this;
+}
+
 void view_curses::mvwattrline(WINDOW *window,
                               int y,
                               int x,
@@ -150,7 +176,7 @@ void view_curses::mvwattrline(WINDOW *window,
 
         tab_iter = tab_list.lower_bound(attr_range.lr_start);
         if (tab_iter != tab_list.end()) {
-            if (attr_range.lr_start > tab_iter->first) {
+            if ((size_t)attr_range.lr_start > tab_iter->first) {
                 attr_range.lr_start += (tab_iter->second - tab_iter->first) - 1;
             }
         }
@@ -158,7 +184,7 @@ void view_curses::mvwattrline(WINDOW *window,
         if (attr_range.lr_end != -1) {
             tab_iter = tab_list.lower_bound(attr_range.lr_end);
             if (tab_iter != tab_list.end()) {
-                if (attr_range.lr_end > tab_iter->first) {
+                if ((size_t)attr_range.lr_end > tab_iter->first) {
                     attr_range.lr_end += (
                         tab_iter->second - tab_iter->first) - 1;
                 }
@@ -226,6 +252,14 @@ void view_curses::mvwattrline(WINDOW *window,
     }
 }
 
+class color_listener : public lnav_config_listener {
+    void reload_config() {
+        view_colors::singleton().init_roles();
+    }
+};
+
+static color_listener _COLOR_LISTENER;
+
 view_colors &view_colors::singleton(void)
 {
     static view_colors s_vc;
@@ -295,7 +329,10 @@ void view_colors::init_roles(void)
 
     /* Setup the mappings from roles to actual colors. */
     this->vc_role_colors[VCR_TEXT]   =
-        ansi_color_pair(COLOR_WHITE, COLOR_BLACK) | A_DIM;
+        ansi_color_pair(COLOR_WHITE, COLOR_BLACK);
+    if (lnav_config.lc_ui_dim_text) {
+        this->vc_role_colors[VCR_TEXT] |= A_DIM;
+    }
     this->vc_role_colors[VCR_SEARCH] =
         this->vc_role_colors[VCR_TEXT] | A_REVERSE;
     this->vc_role_colors[VCR_OK]      = ansi_color_pair(COLOR_GREEN, COLOR_BLACK) | A_BOLD;
@@ -303,6 +340,7 @@ void view_colors::init_roles(void)
     this->vc_role_colors[VCR_WARNING] = ansi_color_pair(COLOR_YELLOW, COLOR_BLACK) | A_BOLD;
     this->vc_role_colors[VCR_ALT_ROW] = ansi_color_pair(COLOR_WHITE, COLOR_BLACK) | A_BOLD;
     this->vc_role_colors[VCR_ADJUSTED_TIME] = ansi_color_pair(COLOR_MAGENTA, COLOR_BLACK);
+    this->vc_role_colors[VCR_SKEWED_TIME] = ansi_color_pair(COLOR_YELLOW, COLOR_BLACK) | A_UNDERLINE;
 
     this->vc_role_colors[VCR_STATUS] =
         ansi_color_pair(COLOR_BLACK, COLOR_WHITE);
@@ -316,6 +354,8 @@ void view_colors::init_roles(void)
         ansi_color_pair(COLOR_GREEN, COLOR_WHITE) | A_BOLD;
     this->vc_role_colors[VCR_BOLD_STATUS] =
         ansi_color_pair(COLOR_BLACK, COLOR_WHITE) | A_BOLD;
+    this->vc_role_colors[VCR_VIEW_STATUS] =
+        ansi_color_pair(COLOR_WHITE, COLOR_BLUE);
 
     this->vc_role_colors[VCR_KEYWORD] = ansi_color_pair(COLOR_BLUE, COLOR_BLACK);
     this->vc_role_colors[VCR_STRING] = ansi_color_pair(COLOR_GREEN, COLOR_BLACK) | A_BOLD;
@@ -326,6 +366,10 @@ void view_colors::init_roles(void)
     this->vc_role_colors[VCR_DIFF_ADD]     = ansi_color_pair(COLOR_GREEN, COLOR_BLACK);
     this->vc_role_colors[VCR_DIFF_SECTION] = ansi_color_pair(COLOR_MAGENTA, COLOR_BLACK);
 
+    this->vc_role_colors[VCR_LOW_THRESHOLD] = ansi_color_pair(COLOR_BLACK, COLOR_GREEN);
+    this->vc_role_colors[VCR_MED_THRESHOLD] = ansi_color_pair(COLOR_BLACK, COLOR_YELLOW);
+    this->vc_role_colors[VCR_HIGH_THRESHOLD] = ansi_color_pair(COLOR_BLACK, COLOR_RED);
+
     for (lpc = 0; lpc < VCR_HIGHLIGHT_START; lpc++) {
         this->vc_role_reverse_colors[lpc] =
             this->vc_role_colors[lpc] | A_REVERSE;
